from erdpy.projects.core import (build_project, clean_project,
                                 get_projects_in_workspace, load_project,
                                 run_tests)
from erdpy.projects.project_base import Project
from erdpy.projects.project_clang import ProjectClang
from erdpy.projects.project_cpp import ProjectCpp
from erdpy.projects.project_rust import ProjectRust
from erdpy.projects.project_sol import ProjectSol
from erdpy.projects.templates import (create_from_template,
                                      list_project_templates)

__all__ = ["build_project", "clean_project", "run_tests", "get_projects_in_workspace", "load_project", "Project", "ProjectClang", "ProjectCpp", "ProjectRust", "ProjectSol", "create_from_template", "list_project_templates"]



import logging
import subprocess
from os import path
from pathlib import Path

from erdpy import dependencies, errors, myprocess, utils
from erdpy.projects.project_base import Project

logger = logging.getLogger("ProjectSol")


class ProjectSol(Project):
    def __init__(self, directory):
        super().__init__(directory)

    def perform_build(self):
        """
        See https://github.com/second-state/SOLL/blob/master/utils/ll2ewasm_sol
        """

        self.unit = self.find_file_globally("*.sol")
        self.unit_name = self.unit.stem
        self.file_ll = self.unit.with_suffix(".ll")
        self.file_functions = self.unit.with_suffix(".functions")
        self.file_main_ll = self.unit.with_suffix(".main.ll")
        self.file_bc = self.unit.with_suffix(".bc")
        self.file_o = self.unit.with_suffix(".o")

        try:
            self._create_main_ll()
            self._emit_LLVM()
            self._emit_funcions()
            self._do_llvm_link()
            self._do_llvm_opt()
            self._do_llc()
            self._do_wasm_ld()
        except subprocess.CalledProcessError as err:
            raise errors.BuildError(err.output)

    def _create_main_ll(self):
        logger.info("_create_main_ll")

        package_path = Path(__file__).parent
        template_path = package_path.joinpath("sol_main_ll.txt")
        template = utils.read_file(template_path)
        content = template.replace("{{NAME}}", self.unit_name)
        utils.write_file(self.file_main_ll, content)

    def _emit_LLVM(self):
        logger.info("_emit_LLVM")

        tool = self._get_soll_path()
        args = [tool, "-action", "EmitLLVM", str(self.unit)]
        output = myprocess.run_process(args)
        utils.write_file(self.file_ll, output)

    def _emit_funcions(self):
        logger.info("_emit_funcions")

        tool = self._get_soll_path()
        args = [tool, "-action", "EmitFuncSig", str(self.unit)]
        output = myprocess.run_process(args)
        utils.write_file(self.file_functions, output)

    def _do_llvm_link(self):
        logger.info("_do_llvm_link")

        tool = path.join(self._get_llvm_path(), "llvm-link")
        args = [tool, self.file_ll, self.file_main_ll, "-o", self.file_bc]
        myprocess.run_process(args)

    def _do_llvm_opt(self):
        logger.info("_do_llvm_opt")

        tool = path.join(self._get_llvm_path(), "opt")
        args = [tool, "-std-link-opts", "-Oz", "-polly", self.file_bc, "-o", self.file_bc]
        myprocess.run_process(args)

    def _do_llc(self):
        logger.info("_do_llc")

        tool = path.join(self._get_llvm_path(), "llc")
        args = [tool, "-O3", "-filetype=obj", self.file_bc, "-o", self.file_o]
        myprocess.run_process(args)

    def _do_wasm_ld(self):
        logger.info("_do_wasm_ld")

        tool = path.join(self._get_llvm_path(), "wasm-ld")
        args = [
            tool,
            "--entry",
            "main",
            "--demangle",
            "--no-gc-sections",
            "--export-all",
            "--allow-undefined",
            "--verbose",
            self.file_o,
            "-o", self.get_file_wasm()
        ]
        myprocess.run_process(args)

    def _get_soll_path(self):
        directory = dependencies.get_module_directory("soll")
        return path.join(directory, "soll")

    def _get_llvm_path(self):
        return dependencies.get_module_directory("llvm")

    def get_dependencies(self):
        return ["soll", "llvm"]



import binascii
from typing import Any
from os import path

from erdpy.validators.validators_file import ValidatorsFile
from erdpy.conv.conv import Converters
from erdpy.accounts import Account
from erdpy.config import MetaChainSystemSCsCost
from erdpy.validators.core import estimate_system_sc_call
from erdpy.wallet.pem import parse_validator_pem
from erdpy.wallet.signing import sign_message_with_bls_key

DELEGATION_MANAGER_SC_ADDRESS = "erd1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqylllslmq6y6"


def prepare_args_for_create_new_staking_contract(args: Any):
    args.data = 'createNewDelegationContract'
    args.data += '@' + Converters.str_int_to_hex_str(str(args.total_delegation_cap))
    args.data += '@' + Converters.str_int_to_hex_str(str(args.service_fee))

    args.receiver = DELEGATION_MANAGER_SC_ADDRESS

    if args.estimate_gas:
        # factor is equals 2 because there 2 delegation manager operations when a contract is created
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_MANAGER_OPS, factor=2)


def prepare_args_for_add_nodes(args: Any):
    validators_file = ValidatorsFile(args.validators_file)

    # TODO: Refactor, so that only address is received here.
    if args.using_delegation_manager:
        account = Account(address=args.delegation_contract)
    elif args.pem:
        account = Account(pem_file=args.pem)
    elif args.keyfile and args.passfile:
        account = Account(key_file=args.keyfile, pass_file=args.passfile)

    add_nodes_data = "addNodes"
    num_of_nodes = validators_file.get_num_of_nodes()
    for validator in validators_file.get_validators_list():
        # get validator
        validator_pem = validator.get("pemFile")
        validator_pem = path.join(path.dirname(args.validators_file), validator_pem)
        seed, bls_key = parse_validator_pem(validator_pem)
        signed_message = sign_message_with_bls_key(account.address.pubkey().hex(), seed.decode('ascii'))
        add_nodes_data += f"@{bls_key}@{signed_message}"

    args.receiver = args.delegation_contract
    args.data = add_nodes_data
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS, num_of_nodes + 1)


def prepare_args_for_remove_nodes(args: Any):
    _prepare_args("removeNodes", args)


def prepare_args_for_stake_nodes(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.bls_keys)
    args.data = 'stakeNodes' + parsed_keys
    args.receiver = args.delegation_contract

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS
                                                 + MetaChainSystemSCsCost.STAKE, num_keys + 1)


def prepare_args_for_unbond_nodes(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.bls_keys)
    args.data = 'unBondNodes' + parsed_keys
    args.receiver = args.delegation_contract

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS
                                                 + MetaChainSystemSCsCost.UNBOND, num_keys + 1)


def prepare_args_for_unstake_nodes(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.bls_keys)
    args.data = 'unStakeNodes' + parsed_keys
    args.receiver = args.delegation_contract

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS
                                                 + MetaChainSystemSCsCost.UNSTAKE, num_keys + 1)


def prepare_args_for_unjail_nodes(args: Any):
    _prepare_args("unJailNodes", args)


def _prepare_args(command: str, args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.bls_keys)
    args.data = command + parsed_keys
    args.receiver = args.delegation_contract

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS, num_keys + 1)


def prepare_args_change_service_fee(args: Any):
    data = 'changeServiceFee'
    data += '@' + Converters.str_int_to_hex_str(str(args.service_fee))

    args.data = data
    args.receiver = args.delegation_contract
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS, 1)


def prepare_args_modify_delegation_cap(args: Any):
    data = 'modifyTotalDelegationCap'
    data += '@' + Converters.str_int_to_hex_str(str(args.delegation_cap))

    args.data = data
    args.receiver = args.delegation_contract
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS, 1)


def prepare_args_automatic_activation(args: Any):
    data = 'setAutomaticActivation'
    if args.set:
        data += '@' + binascii.hexlify(str.encode('true')).decode()

    if args.unset:
        data += '@' + binascii.hexlify(str.encode('false')).decode()

    args.data = data
    args.receiver = args.delegation_contract
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS, 1)


def prepare_args_redelegate_cap(args: Any):
    data = 'setCheckCapOnReDelegateRewards'
    if args.set:
        data += '@' + binascii.hexlify(str.encode('true')).decode()

    if args.unset:
        data += '@' + binascii.hexlify(str.encode('false')).decode()

    args.data = data
    args.receiver = args.delegation_contract
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS, 1)


def prepare_args_set_metadata(args: Any):
    data = 'setMetaData'
    data += '@' + binascii.hexlify(str.encode(args.name)).decode()
    data += '@' + binascii.hexlify(str.encode(args.website)).decode()
    data += '@' + binascii.hexlify(str.encode(args.identifier)).decode()

    args.data = data
    args.receiver = args.delegation_contract
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.DELEGATION_OPS, 1)



from typing import Any

from erdpy.testnet.config import TestnetConfiguration
from erdpy.testnet.nodes_setup_json import CHAIN_ID


def patch(data: Any, testnet_config: TestnetConfiguration):
    # Set chain ID and transaction version
    data['GeneralSettings']['ChainID'] = CHAIN_ID
    data['GeneralSettings']['MinTransactionVersion'] = 1

    data['DbLookupExtensions']['Enabled'] = True
    data['GeneralSettings']['SCDeployEnableEpoch'] = 0
    data['GeneralSettings']['BuiltInFunctionsEnableEpoch'] = 0
    data['GeneralSettings']['RelayedTransactionsEnableEpoch'] = 0
    data['GeneralSettings']['PenalizedTooMuchGasEnableEpoch'] = 0
    data['GeneralSettings']['SwitchJailWaitingEnableEpoch'] = 0
    data['GeneralSettings']['BelowSignedThresholdEnableEpoch'] = 0
    data['GeneralSettings']['SwitchHysteresisForMinNodesEnableEpoch'] = 1
    data['GeneralSettings']['TransactionSignedWithTxHashEnableEpoch'] = 0
    data['GeneralSettings']['MetaProtectionEnableEpoch'] = 0
    data['GeneralSettings']['AheadOfTimeGasUsageEnableEpoch'] = 0
    data['GeneralSettings']['GasPriceModifierEnableEpoch'] = 0

    # Make epochs shorter
    data['EpochStartConfig']['RoundsPerEpoch'] = 100


def patch_api(data: Any, testnet_config: TestnetConfiguration):
    routes = data['APIPackages']['transaction']['Routes']
    for route in routes:
        route["Open"] = True


def patch_system_smart_contracts(data: Any, testnet_config: TestnetConfiguration):
    data['StakingSystemSCConfig']['StakeEnableEpoch'] = 0
    data['StakingSystemSCConfig']['StakingV2Epoch'] = 1
    data['StakingSystemSCConfig']['DoubleKeyProtectionEnableEpoch'] = 0
    data['StakingSystemSCConfig']['ActivateBLSPubKeyMessageVerification'] = True
    data['ESDTSystemSCConfig']['EnabledEpoch'] = 0
    data['GovernanceSystemSCConfig']['EnabledEpoch'] = 0
    data['DelegationManagerSystemSCConfig']['EnabledEpoch'] = 1
    data['DelegationSystemSCConfig']['EnabledEpoch'] = 1



from typing import Any

from erdpy.testnet.config import TestnetConfiguration

PROTOCOL_ID = '/erd/kad/sandbox'


def patch(data: Any, testnet_config: TestnetConfiguration, node_index: int, port_first: int) -> Any:
    data['Node']['Port'] = str(port_first + node_index)
    data['Node']['ThresholdMinConnectedPeers'] = 1
    data['KadDhtPeerDiscovery']['InitialPeerList'] = [
        testnet_config.seednode_address()
    ]
    data['KadDhtPeerDiscovery']['ProtocolID'] = PROTOCOL_ID
    data['Sharding']['Type'] = "NilListSharder"


def patch_for_seednode(data: Any, testnet_config: TestnetConfiguration):
    port_seednode = testnet_config.networking['port_seednode']

    data['Node']['Port'] = str(port_seednode)
    data['Node']['MaximumExpectedPeerCount'] = 16
    data['KadDhtPeerDiscovery']['ProtocolID'] = PROTOCOL_ID
    data['Sharding']['Type'] = "NilListSharder"




from typing import Any, Union


class KnownError(Exception):
    inner = None

    def __init__(self, message: str, inner: Union[Any, None] = None):
        super().__init__(message)
        self.inner = inner

    def get_pretty(self) -> str:
        if self.inner:
            return f"""{self}
... {self.inner}
"""
        return str(self)


class ProgrammingError(KnownError):
    pass


class TemplateMissingError(KnownError):
    def __init__(self, template: str):
        super().__init__(f"Template missing: {template}")


class DownloadError(KnownError):
    pass


class BadUrlError(DownloadError):
    pass


class UnknownArchiveType(KnownError):
    pass


class DependencyMissing(KnownError):
    def __init__(self, name: str, tag: str):
        super().__init__(f"Dependency missing: {name} {tag}")


class UnknownDependency(KnownError):
    def __init__(self, name: str):
        super().__init__(f"Unknown dependency: {name}")


class BadDirectory(KnownError):
    def __init__(self, directory: str):
        super().__init__(f"Bad directory: {directory}")


class BadFile(KnownError):
    def __init__(self, filename: str, inner=None):
        super().__init__(f"Bad file: {filename}.", inner)


class NotSupportedProject(KnownError):
    def __init__(self, directory: str):
        super().__init__(f"Directory is not a supported project: {directory}")


class PlatformNotSupported(KnownError):
    def __init__(self, action_or_item: str, platform: str):
        super().__init__(f"[{action_or_item}] is not supported on platform [{platform}].")


class BuildError(KnownError):
    def __init__(self, message, inner=None):
        super().__init__(f"Build error: {message}.", inner)


class BadSink(ProgrammingError):
    def __init__(self, name: str):
        super().__init__(f"Bad sink:\n {name}.")


class UnknownArgumentFormat(KnownError):
    def __init__(self, argument: Any):
        super().__init__(f"Cannot handle non-hex, non-number arguments yet: {argument}.")


class ProxyRequestError(KnownError):
    def __init__(self, url: str, data: Any):
        super().__init__(f"Proxy request error for url [{url}]: {data}")


class BlockAtlasRequestError(KnownError):
    def __init__(self, url: str, data: Any):
        super().__init__(f"Block Atlas request error for url [{url}]: {data}")


class BadInputError(KnownError):
    def __init__(self, input: str, message: str):
        super().__init__(f"Bad input [{input}]: {message}")


class BadAddressFormatError(KnownError):
    def __init__(self, value: str):
        super().__init__(f"Bad address [{value}].")


class EmptyAddressError(KnownError):
    def __init__(self):
        super().__init__("Address is empty.")


class ExternalProcessError(KnownError):
    def __init__(self, command_line: str, message: str):
        super().__init__(f"""External process error:
Command line: {command_line}
Output: {message}""")


class UnknownConfigurationError(KnownError):
    def __init__(self, name: str):
        super().__init__(f"Configuration entry is not known: {name}.")


class ConfigurationShouldBeUniqueError(KnownError):
    def __init__(self, name: str):
        super().__init__(f"Configuration entry already exists: {name}.")


class ConfigurationProtectedError(KnownError):
    def __init__(self, name: str):
        super().__init__(f"This configuration name is protected: {name}.")


class UnknownDerivationFunction(KnownError):
    def __init__(self):
        super().__init__("Unknown key derivation function.")


class UnknownCipher(KnownError):
    def __init__(self, name: str):
        super().__init__(f"Unknown cipher: {name}.")


class InvalidKeystoreFilePassword(KnownError):
    def __init__(self):
        super().__init__("Provided keystore file password is invalid.")


class BadUserInput(KnownError):
    def __init__(self, message: str):
        super().__init__(f"Bad user input: {message}.")


class BadUsage(KnownError):
    def __init__(self, message: str):
        super().__init__(f"Bad usage: {message}.")


class CannotSignMessageWithBLSKey(KnownError):
    def __init__(self):
        super(CannotSignMessageWithBLSKey, self).__init__("cannot sign message with BLS key")


class CannotReadValidatorsData(KnownError):
    def __init__(self):
        super(CannotReadValidatorsData, self).__init__("cannot read validators data")


class TransactionIsNotSigned(KnownError):
    def __init__(self):
        super().__init__("Transaction is not signed.")


class NoWalletProvided(KnownError):
    def __init__(self):
        super().__init__("No wallet provided.")


class TestnetError(KnownError):
    def __init__(self, message: str):
        super().__init__(message)



from erdpy import cli_shared
import logging
import sys
from typing import Any

from erdpy import config, utils

logger = logging.getLogger("cli.config")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "config", "Configure elrond-sdk (default values etc.)")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "config", "dump", "Dumps configuration.")
    sub.set_defaults(func=dump)

    sub = cli_shared.add_command_subparser(subparsers, "config", "get", "Gets a configuration value.")
    _add_name_arg(sub)
    sub.set_defaults(func=get_value)

    sub = cli_shared.add_command_subparser(subparsers, "config", "set", "Sets a configuration value.")
    _add_name_arg(sub)
    sub.add_argument("value", help="the new value")
    sub.set_defaults(func=set_value)

    sub = cli_shared.add_command_subparser(subparsers, "config", "new", "Creates a new configuration.")
    _add_name_arg(sub)
    sub.add_argument("--template", required=False, help="template from which to create the new config")
    sub.set_defaults(func=new_config)

    sub = cli_shared.add_command_subparser(subparsers, "config", "switch", "Switch to a different config")
    _add_name_arg(sub)
    sub.set_defaults(func=switch_config)

    sub = cli_shared.add_command_subparser(subparsers, "config", "list", "List available configs")
    sub.set_defaults(func=list_configs)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def _add_name_arg(sub: Any):
    sub.add_argument("name", help="the name of the configuration entry")


def dump(args: Any):
    data = config.get_active()
    utils.dump_out_json(data, sys.stdout)


def get_value(args: Any):
    value = config.get_value(args.name)
    print(value)


def set_value(args: Any):
    config.set_value(args.name, args.value)


def new_config(args: Any):
    config.create_new_config(name=args.name, template=args.template)
    dump(None)


def switch_config(args: Any):
    config.set_active(args.name)
    dump(None)


def list_configs(args: Any):
    data = config.read_file()
    configurations = data.get('configurations', {})
    for config_name in configurations.keys():
        if config_name == data.get("active", "default"):
            config_name += "*"
        print(config_name)



import logging
import os
from typing import Any

from erdpy import cli_shared, errors, projects, utils
from erdpy.accounts import Account, Address
from erdpy.contracts import CodeMetadata, SmartContract
from erdpy.projects import load_project
from erdpy.proxy.core import ElrondProxy
from erdpy.transactions import Transaction

logger = logging.getLogger("cli.contracts")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "contract", "Build, deploy and interact with Smart Contracts")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "contract", "new", "Create a new Smart Contract project based on a template.")
    sub.add_argument("name")
    sub.add_argument("--template", required=True, help="the template to use")
    sub.add_argument("--directory", type=str, default=os.getcwd(), help="🗀 the parent directory of the project (default: current directory)")
    sub.set_defaults(func=create)

    sub = cli_shared.add_command_subparser(subparsers, "contract", "templates", "List the available Smart Contract templates.")
    sub.set_defaults(func=list_templates)

    sub = cli_shared.add_command_subparser(subparsers, "contract", "build", "Build a Smart Contract project using the appropriate buildchain.")
    _add_project_arg(sub)
    sub.add_argument("--debug", action="store_true", default=False, help="set debug flag (default: %(default)s)")
    sub.add_argument("--no-optimization", action="store_true", default=False, help="bypass optimizations (for clang) (default: %(default)s)")
    sub.add_argument("--cargo-target-dir", type=str, help="for rust projects, forward the parameter to Cargo")
    sub.add_argument("--wasm-symbols", action="store_true", default=False, help="for rust projects, does not strip the symbols from the wasm output. Useful for analysing the bytecode. Creates larger wasm files. Avoid in production (default: %(default)s)")
    sub.add_argument("--wasm-name", type=str, help="for rust projects, optionally specify the name of the wasm bytecode output file")
    sub.set_defaults(func=build)

    sub = cli_shared.add_command_subparser(subparsers, "contract", "clean", "Clean a Smart Contract project.")
    _add_project_arg(sub)
    sub.set_defaults(func=clean)

    sub = cli_shared.add_command_subparser(subparsers, "contract", "test", "Run Mandos tests.")
    _add_project_arg(sub)
    sub.add_argument("--directory", default="mandos", help="🗀 the directory containing the tests (default: %(default)s)")
    sub.add_argument("--wildcard", required=False, help="wildcard to match only specific test files")
    sub.set_defaults(func=run_tests)

    sub = cli_shared.add_command_subparser(subparsers, "contract", "deploy", "Deploy a Smart Contract.")
    _add_project_or_bytecode_arg(sub)
    _add_metadata_arg(sub)
    cli_shared.add_outfile_arg(sub)
    cli_shared.add_wallet_args(sub)
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_tx_args(sub, with_receiver=False, with_data=False)
    _add_arguments_arg(sub)
    cli_shared.add_broadcast_args(sub)

    sub.set_defaults(func=deploy)

    sub = cli_shared.add_command_subparser(subparsers, "contract", "call", "Interact with a Smart Contract (execute function).")
    _add_contract_arg(sub)
    cli_shared.add_outfile_arg(sub)
    cli_shared.add_wallet_args(sub)
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_tx_args(sub, with_receiver=False, with_data=False)
    _add_function_arg(sub)
    _add_arguments_arg(sub)
    sub.add_argument("--wait-result", action="store_true", default=False,
                     help="signal to wait for the transaction result - only valid if --send is set")
    sub.add_argument("--timeout", default=100, help="max num of seconds to wait for result"
                                                    " - only valid if --wait-result is set")
    cli_shared.add_broadcast_args(sub, relay=True)

    sub.set_defaults(func=call)

    sub = cli_shared.add_command_subparser(subparsers, "contract", "upgrade", "Upgrade a previously-deployed Smart Contract")
    _add_contract_arg(sub)
    cli_shared.add_outfile_arg(sub)
    _add_project_or_bytecode_arg(sub)
    _add_metadata_arg(sub)
    cli_shared.add_wallet_args(sub)
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_tx_args(sub, with_receiver=False, with_data=False)
    _add_arguments_arg(sub)
    cli_shared.add_broadcast_args(sub)

    sub.set_defaults(func=upgrade)

    sub = cli_shared.add_command_subparser(subparsers, "contract", "query", "Query a Smart Contract (call a pure function)")
    _add_contract_arg(sub)
    cli_shared.add_proxy_arg(sub)
    _add_function_arg(sub)
    _add_arguments_arg(sub)
    sub.set_defaults(func=query)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def _add_project_arg(sub: Any):
    sub.add_argument("project", nargs='?', default=os.getcwd(), help="🗀 the project directory (default: current directory)")


def _add_project_or_bytecode_arg(sub: Any):
    group = sub.add_mutually_exclusive_group(required=True)
    group.add_argument("--project", default=os.getcwd(), help="🗀 the project directory (default: current directory)")
    group.add_argument("--bytecode", help="the WASM file")


def _add_contract_arg(sub: Any):
    sub.add_argument("contract", help="🖄 the address of the Smart Contract")


def _add_function_arg(sub: Any):
    sub.add_argument("--function", required=True, help="the function to call")


def _add_arguments_arg(sub: Any):
    sub.add_argument("--arguments", nargs='+', help="arguments for the contract transaction, as numbers or hex-encoded. E.g. --arguments 42 0x64 1000 0xabba")


def _add_metadata_arg(sub: Any):
    sub.add_argument("--metadata-not-upgradeable", dest="metadata_upgradeable", action="store_false", help="‼ mark the contract as NOT upgradeable (default: upgradeable)")
    sub.add_argument("--metadata-payable", dest="metadata_payable", action="store_true", help="‼ mark the contract as payable (default: not payable)")
    sub.set_defaults(metadata_upgradeable=True, metadata_payable=False)


def list_templates(args: Any):
    projects.list_project_templates()


def create(args: Any):
    name = args.name
    template = args.template
    directory = args.directory

    projects.create_from_template(name, template, directory)


def clean(args: Any):
    project = args.project
    projects.clean_project(project)


def build(args: Any):
    project = args.project
    options = {
        "debug": args.debug,
        "optimized": not args.no_optimization,
        "verbose": args.verbose,
        "cargo_target_dir": args.cargo_target_dir,
        "wasm_symbols": args.wasm_symbols,
        "wasm_name": args.wasm_name
    }

    projects.build_project(project, options)


def run_tests(args: Any):
    projects.run_tests(args)


def deploy(args: Any):
    logger.debug("deploy")
    cli_shared.check_broadcast_args(args)

    arguments = args.arguments
    gas_price = args.gas_price
    gas_limit = args.gas_limit
    value = args.value
    chain = args.chain
    version = args.version

    contract = _prepare_contract(args)
    sender = _prepare_sender(args)

    tx = contract.deploy(sender, arguments, gas_price, gas_limit, value, chain, version)
    logger.info("Contract address: %s", contract.address)

    result = None
    try:
        result = _send_or_simulate(tx, args)
    finally:
        txdict = tx.to_dump_dict()
        txdict['address'] = contract.address.bech32()
        dump_tx_and_result(txdict, result, args)


def dump_tx_and_result(tx: Any, result: Any, args: Any):
    dump = dict()
    dump['emitted_tx'] = tx

    try:
        returnCode = ''
        returnMessage = ''
        dump['result'] = result['result']
        for scrHash, scr in dump['result']['scResults'].items():
            if scr['receiver'] == tx['tx']['sender']:
                retCode = scr['data'][1:]
                retCode = bytes.fromhex(retCode).decode('ascii')
                returnCode = retCode
                returnMessage = scr['returnMessage']

        dump['result']['returnCode'] = returnCode
        dump['result']['returnMessage'] = returnMessage
    except TypeError:
        pass

    utils.dump_out_json(dump, outfile=args.outfile)


def _prepare_contract(args: Any) -> SmartContract:
    if args.bytecode:
        bytecode = utils.read_file(args.bytecode, binary=True).hex()
    else:
        project = load_project(args.project)
        bytecode = project.get_bytecode()

    metadata = CodeMetadata(args.metadata_upgradeable, args.metadata_payable)
    contract = SmartContract(bytecode=bytecode, metadata=metadata)
    return contract


def _prepare_sender(args: Any) -> Account:
    if args.pem:
        sender = Account(pem_file=args.pem, pem_index=args.pem_index)
    elif args.keyfile and args.passfile:
        sender = Account(key_file=args.keyfile, pass_file=args.passfile)
    else:
        raise errors.NoWalletProvided()

    sender.nonce = args.nonce
    if args.recall_nonce:
        sender.sync_nonce(ElrondProxy(args.proxy))

    return sender


def call(args: Any):
    logger.debug("call")
    cli_shared.check_broadcast_args(args)

    contract_address = args.contract
    function = args.function
    arguments = args.arguments
    gas_price = args.gas_price
    gas_limit = args.gas_limit
    value = args.value
    chain = args.chain
    version = args.version

    contract = SmartContract(contract_address)
    sender = _prepare_sender(args)

    tx = contract.execute(sender, function, arguments, gas_price, gas_limit, value, chain, version)
    try:
        result = _send_or_simulate(tx, args)
    finally:
        txdict = tx.to_dump_dict()
        dump_tx_and_result(txdict, result, args)


def upgrade(args: Any):
    logger.debug("upgrade")
    cli_shared.check_broadcast_args(args)

    contract_address = args.contract
    arguments = args.arguments
    gas_price = args.gas_price
    gas_limit = args.gas_limit
    value = args.value
    chain = args.chain
    version = args.version

    contract = _prepare_contract(args)
    contract.address = Address(contract_address)
    sender = _prepare_sender(args)

    tx = contract.upgrade(sender, arguments, gas_price, gas_limit, value, chain, version)
    try:
        result = _send_or_simulate(tx, args)
    finally:
        txdict = tx.to_dump_dict()
        dump_tx_and_result(txdict, result, args)


def query(args: Any):
    logger.debug("query")

    contract_address = args.contract
    function = args.function
    arguments = args.arguments

    contract = SmartContract(contract_address)
    result = contract.query(ElrondProxy(args.proxy), function, arguments)
    utils.dump_out_json(result)


def _send_or_simulate(tx: Transaction, args: Any):
    send_wait_result = args.wait_result and args.send and not args.simulate
    send_only = args.send and not (args.wait_result or args.simulate)
    simulate = args.simulate and not (send_only or send_wait_result)

    if send_wait_result:
        proxy = ElrondProxy(args.proxy)
        response = tx.send_wait_result(proxy, args.timeout)
        return None
    elif send_only:
        tx.send(ElrondProxy(args.proxy))
        return None
    elif simulate:
        response = tx.simulate(ElrondProxy(args.proxy))
        return response



import asyncio
import logging
import subprocess
import traceback
from typing import Any, List

from erdpy import errors

logger = logging.getLogger("myprocess")


def run_process(args: List[str], env: Any = None, dump_to_stdout: bool = True, cwd: str = None):
    logger.info(f"run_process: {args}, in folder: {cwd}")

    try:
        output = subprocess.check_output(args, shell=False, universal_newlines=True, stderr=subprocess.STDOUT, env=env, cwd=cwd)
        logger.info("Successful run. Output:")
        if dump_to_stdout:
            print(output or "[No output]")
        return output
    except subprocess.CalledProcessError as error:
        raise errors.ExternalProcessError(error.cmd, error.output)


def run_process_async(args: List[str], env: Any = None, cwd: str = None, stdout_sink=None, stderr_sink=None):
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(_async_subprocess(args, env, stdout_sink, stderr_sink, cwd))
    loop.close()
    asyncio.set_event_loop(asyncio.new_event_loop())
    return result


async def _async_subprocess(args, env, stdout_sink, stderr_sink, cwd: str):
    process = await asyncio.create_subprocess_exec(*args, env=env, stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.PIPE, cwd=cwd)

    await asyncio.wait([
        _read_stream(process.stdout, stdout_sink),
        _read_stream(process.stderr, stderr_sink)
    ])
    return await process.wait()


async def _read_stream(stream, sink):
    sink = sink or ConsoleOutputSink()
    sink.open()

    while True:
        try:
            line = await stream.readline()
            if line:
                line = line.decode("utf-8", "replace").strip()
                sink.write(line)
            else:
                break
        except Exception:
            print(traceback.format_exc())

    sink.close()


class OutputSink:
    def open(self):
        pass

    def write(self, line: str):
        pass

    def close(self):
        pass


class ConsoleOutputSink(OutputSink):
    def write(self, line: str):
        print(line)


class FileOutputSink(OutputSink):
    def __init__(self, filepath: str) -> None:
        super().__init__()
        self.filepath = filepath
        self.file = None

    def open(self):
        self.file = open(self.filepath, "w")

    def write(self, line: str):
        if self.file is None:
            return

        self.file.write(line)

    def close(self):
        if self.file is None:
            return

        self.file.close()



import base64
import itertools
import textwrap
from os import path
from pathlib import Path
from typing import Tuple, Union

from erdpy import guards, utils


def get_pubkey(pem_file: Union[str, Path]):
    _, pubkey = parse(pem_file)
    return pubkey


def parse(pem_file: Union[str, Path], index: int = 0) -> Tuple[bytes, bytes]:
    pem_file = path.expanduser(pem_file)
    guards.is_file(pem_file)

    lines = utils.read_lines(pem_file)
    keys_lines = [list(key_lines) for is_next_key, key_lines in itertools.groupby(lines, lambda line: "-----" in line)
                  if not is_next_key]
    keys = ["".join(key_lines) for key_lines in keys_lines]

    key_base64 = keys[index]
    key_hex = base64.b64decode(key_base64).decode()
    key_bytes = bytes.fromhex(key_hex)

    seed = key_bytes[:32]
    pubkey = key_bytes[32:]
    return seed, pubkey


def parse_validator_pem(pem_file, index: int = 0):
    pem_file = path.expanduser(pem_file)
    guards.is_file(pem_file)

    lines = utils.read_lines(pem_file)
    bls_keys = read_bls_keys(lines)
    private_keys = read_validators_private_keys(lines)

    private_key = private_keys[index]
    key_bytes = get_bytes_from_private_key(private_key)

    bls_key = bls_keys[index]
    seed = key_bytes
    return seed, bls_key


def read_bls_keys(lines):
    bls_keys = []

    for line in lines:
        splited_line = line.split(" ")
        if len(splited_line) < 5:
            continue
        if "BEGIN" in splited_line[0]:
            continue

        token = splited_line[4].replace('-----', '')
        bls_keys.append(token)

    return bls_keys


def read_validators_private_keys(lines):
    private_keys = []

    private_keys_lines = [list(key_lines) for is_next_key, key_lines in
                          itertools.groupby(lines, lambda line: "-----" in line) if not is_next_key]
    for key_list in private_keys_lines:
        private_keys.append(key_list[0] + key_list[1])

    return private_keys


def get_bytes_from_private_key(private_key):
    key_base64 = private_key
    key_hex = base64.b64decode(key_base64).hex()
    key_bytes = bytes.fromhex(key_hex)

    return key_bytes


def write(pem_file: Union[str, Path], seed: bytes, pubkey: bytes, name: str = ""):
    pem_file = path.expanduser(pem_file)

    if not name:
        name = pubkey.hex()

    header = f"-----BEGIN PRIVATE KEY for {name}-----"
    footer = f"-----END PRIVATE KEY for {name}-----"

    seed_hex = seed.hex()
    pubkey_hex = pubkey.hex()
    combined = seed_hex + pubkey_hex
    combined = combined.encode()
    key_base64 = base64.b64encode(combined).decode()

    payload_lines = textwrap.wrap(key_base64, 64)
    payload = "\n".join(payload_lines)
    content = "\n".join([header, payload, footer])
    utils.write_file(pem_file, content)



import bottle
from bottle import run, template, static_file, request
from pathlib import Path
from erdpy.projects import get_projects_in_workspace
from erdpy.ide.view_models import SmartContractViewModel
from erdpy.ide.bottle_plugins import EnableCors

app = bottle.Bottle()
app.install(EnableCors())
global_workspace = "./examples/contracts"


@app.route("/")
def index():
    return template("dashboard.html")


@app.route("/workspace/contracts")
def workspace_contracts():
    projects = get_projects_in_workspace(global_workspace)
    models = [SmartContractViewModel(project).__dict__ for project in projects]
    return dict(data=models)


def load_workspace_contracts():
    pass


@app.post("/contracts/deploy")
def contracts_deploy():
    try:
        data = request.json
        print(data)
    except Exception:
        raise ValueError


@app.route('/static/<filepath:path>')
def server_static(filepath):
    return static_file(filepath, root=_get_ide_folder("static"))


def run_ide(workspace):
    global global_workspace
    global_workspace = workspace
    run(app, host="localhost", port=8081, debug=True)


def _get_ide_folder(subfolder):
    return Path(__file__).parent.joinpath(subfolder)


bottle.TEMPLATE_PATH.append(_get_ide_folder("views"))



from erdpy.delegation.staking_provider import prepare_args_modify_delegation_cap, \
    prepare_args_for_unjail_nodes, \
    prepare_args_for_stake_nodes, \
    prepare_args_for_add_nodes, \
    prepare_args_for_unstake_nodes, \
    prepare_args_for_create_new_staking_contract, \
    prepare_args_for_unbond_nodes, \
    prepare_args_for_remove_nodes, \
    prepare_args_change_service_fee, \
    prepare_args_automatic_activation, \
    prepare_args_set_metadata

__all__ = ["prepare_args_modify_delegation_cap",
           "prepare_args_for_unjail_nodes",
           "prepare_args_for_stake_nodes",
           "prepare_args_for_add_nodes",
           "prepare_args_for_unstake_nodes",
           "prepare_args_for_create_new_staking_contract",
           "prepare_args_for_unbond_nodes",
           "prepare_args_for_remove_nodes",
           "prepare_args_change_service_fee",
           "prepare_args_automatic_activation",
           "prepare_args_set_metadata"]



import logging
import traceback

from erdpy import errors
from erdpy.contracts import SmartContract
from erdpy.proxy import ElrondProxy

logger = logging.getLogger("environments")


class Environment:
    def __init__(self):
        pass

    def run_flow(self):
        raise NotImplementedError()

    def deploy_contract(self, contract, owner, arguments, gas_price, gas_limit, chain, version):
        raise NotImplementedError()

    def execute_contract(self, contract, caller, function, arguments, gas_price, gas_limit, chain, version):
        raise NotImplementedError()

    def query_contract(self, contract, function, arguments=None):
        raise NotImplementedError()


class TestnetEnvironment(Environment):
    def __init__(self, url):
        super().__init__()
        self.url = url

    def run_flow(self, flow):
        return self._wrap_flow(flow)

    def _wrap_flow(self, flow):
        try:
            logger.debug("Starting flow...")
            result = flow()
            logger.debug("Flow ran.")
            return result
        except errors.KnownError as err:
            logger.critical(err)
        except Exception:
            print(traceback.format_exc())

    def deploy_contract(self, contract: SmartContract, owner, arguments, gas_price, gas_limit, value, chain, version):
        logger.debug("deploy_contract")
        tx = contract.deploy(owner, arguments, gas_price, gas_limit, value, chain, version)
        proxy = self._get_proxy()
        tx_hash = tx.send(proxy)
        return tx_hash, contract.address

    def execute_contract(self, contract: SmartContract, caller, function, arguments, gas_price, gas_limit, value, chain, version):
        logger.debug("execute_contract: %s", contract.address.bech32())
        tx = contract.execute(caller, function, arguments, gas_price, gas_limit, value, chain, version)
        proxy = self._get_proxy()
        tx_hash = tx.send(proxy)
        return tx_hash

    def upgrade_contract(self, contract: SmartContract, caller, arguments, gas_price, gas_limit, value, chain, version):
        logger.debug("upgrade_contract: %s", contract.address.bech32())
        tx = contract.upgrade(caller, arguments, gas_price, gas_limit, value, chain, version)
        proxy = self._get_proxy()
        tx_hash = tx.send(proxy)
        return tx_hash

    def query_contract(self, contract, function, arguments=None):
        logger.debug("query_contract: %s", contract.address.bech32())
        proxy = self._get_proxy()
        return_data = contract.query(proxy, function, arguments)
        return return_data

    def _get_proxy(self):
        return ElrondProxy(self.url)



import argparse
import logging
import sys
from argparse import ArgumentParser
from typing import Any, List

import erdpy.cli_accounts
import erdpy.cli_block
import erdpy.cli_blockatlas
import erdpy.cli_config
import erdpy.cli_contracts
import erdpy.cli_cost
import erdpy.cli_data
import erdpy.cli_deps
import erdpy.cli_dispatcher
import erdpy.cli_network
import erdpy.cli_testnet
import erdpy.cli_transactions
import erdpy.cli_validators
import erdpy.cli_wallet
import erdpy.cli_delagation
import erdpy.cli_dns
from erdpy import errors, scope
from erdpy._version import __version__

logger = logging.getLogger("cli")


def main():
    try:
        _do_main()
    except errors.KnownError as err:
        logger.critical(err.get_pretty())
        sys.exit(1)
    except KeyboardInterrupt:
        print("erdpy process killed by user.")
        sys.exit(1)


def _do_main():
    logging.basicConfig(level=logging.INFO)
    scope.initialize()

    parser = setup_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARN)

    if not hasattr(args, "func"):
        parser.print_help()
    else:
        args.func(args)


def setup_parser():
    parser = ArgumentParser(
        prog="erdpy",
        usage="erdpy [-h] [-v] [--verbose] COMMAND-GROUP [-h] COMMAND ...",
        description="""
-----------
DESCRIPTION
-----------
erdpy is part of the elrond-sdk and consists of Command Line Tools and Python SDK
for interacting with the Blockchain (in general) and with Smart Contracts (in particular).

erdpy targets a broad audience of users and developers.
https://docs.elrond.com/tools/erdpy.
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser._positionals.title = "COMMAND GROUPS"
    parser._optionals.title = "TOP-LEVEL OPTIONS"

    parser.add_argument("-v", "--version", action="version", version=f"erdpy {__version__}")
    parser.add_argument("--verbose", action="store_true", default=False)

    subparsers = parser.add_subparsers()
    commands: List[Any] = []

    commands.append(erdpy.cli_contracts.setup_parser(subparsers))
    commands.append(erdpy.cli_transactions.setup_parser(subparsers))
    commands.append(erdpy.cli_validators.setup_parser(subparsers))
    commands.append(erdpy.cli_accounts.setup_parser(subparsers))
    commands.append(erdpy.cli_wallet.setup_parser(subparsers))
    commands.append(erdpy.cli_network.setup_parser(subparsers))
    commands.append(erdpy.cli_cost.setup_parser(subparsers))
    commands.append(erdpy.cli_dispatcher.setup_parser(subparsers))
    commands.append(erdpy.cli_blockatlas.setup_parser(subparsers))
    commands.append(erdpy.cli_deps.setup_parser(subparsers))
    commands.append(erdpy.cli_config.setup_parser(subparsers))
    commands.append(erdpy.cli_block.setup_parser(subparsers))
    commands.append(erdpy.cli_testnet.setup_parser(subparsers))
    commands.append(erdpy.cli_data.setup_parser(subparsers))
    commands.append(erdpy.cli_delagation.setup_parser(subparsers))
    commands.append(erdpy.cli_dns.setup_parser(subparsers))

    parser.epilog = """
----------------------
COMMAND GROUPS summary
----------------------
"""
    for choice, sub in subparsers.choices.items():
        parser.epilog += (f"{choice.ljust(30)} {sub.description}\n")

    return parser


if __name__ == "__main__":
    main()



from erdpy import dependencies
import logging
from os import path
from typing import Any, Dict

from erdpy import errors, utils, guards
from erdpy.projects import shared
from erdpy.projects.project_clang import ProjectClang
from erdpy.projects.project_cpp import ProjectCpp
from erdpy.projects.project_rust import ProjectRust
from erdpy.projects.project_sol import ProjectSol

logger = logging.getLogger("projects.core")


def load_project(directory):
    guards.is_directory(directory)

    if shared.is_source_clang(directory):
        return ProjectClang(directory)
    if shared.is_source_cpp(directory):
        return ProjectCpp(directory)
    if shared.is_source_sol(directory):
        return ProjectSol(directory)
    if shared.is_source_rust(directory):
        return ProjectRust(directory)
    else:
        raise errors.NotSupportedProject(directory)


def build_project(directory: str, options: Dict[str, Any]):
    directory = path.expanduser(directory)

    logger.info("build_project.directory: %s", directory)
    logger.info("build_project.debug: %s", options['debug'])

    guards.is_directory(directory)
    project = load_project(directory)
    project.build(options)
    logger.info("Build ran.")


def clean_project(directory: str):
    directory = path.expanduser(directory)
    guards.is_directory(directory)
    project = load_project(directory)
    project.clean()
    logger.info("Project cleaned.")


def run_tests(args: Any):
    project = args.project
    directory = args.directory
    wildcard = args.wildcard

    logger.info("run_tests.project: %s", project)

    dependencies.install_module("arwentools")

    guards.is_directory(project)
    project = load_project(project)
    project.run_tests(directory, wildcard)


def get_projects_in_workspace(workspace):
    guards.is_directory(workspace)
    subfolders = utils.get_subfolders(workspace)
    projects = []

    for folder in subfolders:
        project_directory = path.join(workspace, folder)

        try:
            project = load_project(project_directory)
            projects.append(project)
        except Exception:
            pass

    return projects



from erdpy.proxy.core import ElrondProxy
from erdpy import cli_shared
import logging
from typing import Any

logger = logging.getLogger("cli.network")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "network", "Get Network parameters, such as number of shards, chain identifier etc.")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "network", "num-shards", "Get the number of shards.")
    cli_shared.add_proxy_arg(sub)
    sub.set_defaults(func=get_num_shards)

    sub = cli_shared.add_command_subparser(subparsers, "network", "block-nonce", "Get the latest block nonce, by shard.")
    cli_shared.add_proxy_arg(sub)
    sub.add_argument("--shard", required=True, help="the shard ID (use 4294967295 for metachain)")
    sub.set_defaults(func=get_last_block_nonce)

    sub = cli_shared.add_command_subparser(subparsers, "network", "chain", "Get the chain identifier.")
    cli_shared.add_proxy_arg(sub)
    sub.set_defaults(func=get_chain_id)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def get_num_shards(args):
    proxy_url = args.proxy
    proxy = ElrondProxy(proxy_url)
    num_shards = proxy.get_num_shards()
    print(num_shards)
    return num_shards


def get_last_block_nonce(args):
    proxy_url = args.proxy
    shard = args.shard
    proxy = ElrondProxy(proxy_url)
    nonce = proxy.get_last_block_nonce(shard)
    print(nonce)
    return nonce


def get_chain_id(args):
    proxy_url = args.proxy
    proxy = ElrondProxy(proxy_url)
    chain_id = proxy.get_chain_id()
    print(chain_id)
    return chain_id



import shutil
from collections import OrderedDict
from pathlib import Path
from typing import Dict, Tuple

from erdpy import errors, utils
from erdpy.accounts import Account
from erdpy.wallet import pem

MAX_NUM_NODES = 12


def copy_all_to(destination: str):
    shutil.copytree(_get_folder(), destination)


def copy_validator_key_to(validator_index: int, destination: str):
    shutil.copy(get_validator_key_file(validator_index), destination)


def get_validator_key_file(validator_index: int):
    _guard_validator_index(validator_index)
    return _get_validators_folder().joinpath("validatorKey{:02}.pem".format(validator_index))


def get_observer_key_file(observer_index: int):
    _guard_validator_index(observer_index)
    return _get_observers_folder().joinpath("observerKey{:02}.pem".format(observer_index))


def get_validator_wallets(num_validators: int) -> Dict[str, Account]:
    result = {}

    for i in range(0, num_validators):
        pem_file = get_validator_wallet_file(i)
        nickname = "validator{:02}".format(i)
        account = Account(pem_file=pem_file)
        result[nickname] = account

    return result


def get_validators(num_validators: int) -> Dict[str, Tuple[str, Account]]:
    result = {}

    for i in range(0, num_validators):
        validator_pem_file = get_validator_key_file(i)
        _, pubkey = pem.parse_validator_pem(validator_pem_file)

        pem_file = get_validator_wallet_file(i)
        nickname = "validator{:02}".format(i)
        account = Account(pem_file=pem_file)
        result[nickname] = [pubkey, account]

    return result


def get_validator_wallet_file(validator_index: int):
    _guard_validator_index(validator_index)
    return _get_validators_folder().joinpath("wallet{:02}.pem".format(validator_index))


def _guard_validator_index(validator_index: int):
    if validator_index >= MAX_NUM_NODES:
        raise errors.TestnetError(f"Invalid validator index: {validator_index} >= {MAX_NUM_NODES}.")


def _get_validators_folder():
    return _get_folder().joinpath("validators")


def _get_observers_folder():
    return _get_folder().joinpath("observers")


def get_users() -> Dict[str, Account]:
    result = OrderedDict()

    for pem_file in sorted(utils.list_files(_get_users_folder(), ".pem")):
        nickname = Path(pem_file).stem
        account = Account(pem_file=pem_file)
        result[nickname] = account

    return result


def _get_users_folder():
    return _get_folder().joinpath("users")


def _get_folder():
    return Path(__file__).parent.absolute().joinpath("wallets")



import logging
from typing import Any

from erdpy import cli_shared, utils
from erdpy.accounts import Address
from erdpy.proxy.core import ElrondProxy

logger = logging.getLogger("cli.accounts")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "account", "Get Account data (nonce, balance) from the Network")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "account", "get", "Query account details (nonce, balance etc.)")
    cli_shared.add_proxy_arg(sub)
    _add_address_arg(sub)
    mutex = sub.add_mutually_exclusive_group()
    mutex.add_argument("--balance", action="store_true", help="whether to only fetch the balance")
    mutex.add_argument("--nonce", action="store_true", help="whether to only fetch the nonce")
    mutex.add_argument("--username", action="store_true", help="whether to only fetch the username")
    cli_shared.add_omit_fields_arg(sub)
    sub.set_defaults(func=get_account)

    sub = cli_shared.add_command_subparser(subparsers, "account", "get-transactions", "Query account transactions")
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_outfile_arg(sub)
    _add_address_arg(sub)
    sub.set_defaults(func=get_account_transactions)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def _add_address_arg(sub: Any):
    sub.add_argument("--address", required=True, help="🖄 the address to query")


def get_account(args: Any):
    proxy_url = args.proxy
    address = args.address
    proxy = ElrondProxy(proxy_url)
    account = proxy.get_account(Address(address))
    omit_fields = cli_shared.parse_omit_fields_arg(args)

    if args.balance:
        print(account.get("balance", 0))
    elif args.nonce:
        print(account.get("nonce", 0))
    elif args.username:
        print(account.get("username", 0))
    else:
        utils.omit_fields(account, omit_fields)
        utils.dump_out_json(account)


def get_account_transactions(args: Any):
    proxy_url = args.proxy
    address = args.address
    proxy = ElrondProxy(proxy_url)

    response = proxy.get_account_transactions(Address(address))
    utils.dump_out_json(response, args.outfile)



import bottle
from bottle import response

class EnableCors(object):
    """
    https://stackoverflow.com/a/17262900
    """

    name = "enable_cors"
    api = 2

    def apply(self, fn, context):
        def _enable_cors(*args, **kwargs):
            # set CORS headers
            response.headers["Access-Control-Allow-Origin"] = "*"
            response.headers["Access-Control-Allow-Methods"] = "GET, POST, PUT, OPTIONS"
            response.headers["Access-Control-Allow-Headers"] = "Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token"

            if bottle.request.method != "OPTIONS":
                # actual request; reply with the actual response
                return fn(*args, **kwargs)

        return _enable_cors


from argparse import FileType
from typing import Any

from erdpy import cli_shared, utils
from erdpy.proxy.core import ElrondProxy
from erdpy.transactions import Transaction, do_prepare_transaction


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "tx", "Create and broadcast Transactions")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "tx", "new", "Create a new transaction")
    _add_common_arguments(sub)
    cli_shared.add_outfile_arg(sub, what="signed transaction, hash")
    cli_shared.add_broadcast_args(sub, relay=True)
    cli_shared.add_proxy_arg(sub)
    sub.add_argument("--wait-result", action="store_true", default=False,
                     help="signal to wait for the transaction result - only valid if --send is set")
    sub.add_argument("--timeout", default=100, help="max num of seconds to wait for result"
                                                    " - only valid if --wait-result is set")
    sub.set_defaults(func=create_transaction)

    sub = cli_shared.add_command_subparser(subparsers, "tx", "send", "Send a previously saved transaction")
    cli_shared.add_infile_arg(sub, what="a previously saved transaction")
    cli_shared.add_outfile_arg(sub, what="the hash")
    cli_shared.add_proxy_arg(sub)
    sub.set_defaults(func=send_transaction)

    sub = cli_shared.add_command_subparser(subparsers, "tx", "get", "Get a transaction")
    sub.add_argument("--hash", required=True, help="the hash")
    sub.add_argument("--sender", required=False, help="the sender address")
    sub.add_argument("--with-results", action="store_true", help="will also return the results of transaction")
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_omit_fields_arg(sub)
    sub.set_defaults(func=get_transaction)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def _add_common_arguments(sub: Any):
    cli_shared.add_wallet_args(sub)
    cli_shared.add_tx_args(sub)
    sub.add_argument("--data-file", type=FileType("r"), default=None, help="a file containing transaction data")


def create_transaction(args: Any):
    args = utils.as_object(args)

    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)

    if args.data_file:
        args.data = utils.read_file(args.data_file)

    tx = do_prepare_transaction(args)

    if hasattr(args, "relay") and args.relay:
        args.outfile.write(tx.serialize_as_inner())
        return

    send_wait_result = args.wait_result and args.send and not args.simulate
    send_only = args.send and not (args.wait_result or args.simulate)
    simulate = args.simulate and not (send_only or send_wait_result)

    try:
        if send_wait_result:
            proxy = ElrondProxy(args.proxy)
            response = tx.send_wait_result(proxy, args.timeout)
            utils.dump_out_json(response)
        elif send_only:
            tx.send(ElrondProxy(args.proxy))
        elif simulate:
            response = tx.simulate(ElrondProxy(args.proxy))
            utils.dump_out_json(response)
    finally:
        tx.dump_to(args.outfile)


def send_transaction(args: Any):
    args = utils.as_object(args)

    tx = Transaction.load_from_file(args.infile)

    try:
        tx.send(ElrondProxy(args.proxy))
    finally:
        tx.dump_to(args.outfile)


def get_transaction(args: Any):
    args = utils.as_object(args)
    omit_fields = cli_shared.parse_omit_fields_arg(args)

    proxy = ElrondProxy(args.proxy)

    transaction = proxy.get_transaction(args.hash, args.sender, args.with_results)
    utils.omit_fields(transaction, omit_fields)
    utils.dump_out_json(transaction)



import os.path
from typing import Any, Dict

from erdpy import errors, utils

ROOT_FOLDER_NAME = "elrondsdk"
CONFIG_PATH = os.path.expanduser("~/elrondsdk/erdpy.json")

DEFAULT_GAS_PRICE = 1000000000
GAS_PER_DATA_BYTE = 1500
MIN_GAS_LIMIT = 50000


class MetaChainSystemSCsCost:
    STAKE = 5000000
    UNSTAKE = 5000000
    UNBOND = 5000000
    CLAIM = 5000000
    GET = 5000000
    CHANGE_REWARD_ADDRESS = 5000000
    CHANGE_VALIDATOR_KEYS = 5000000
    UNJAIL = 5000000
    DELEGATION_MANAGER_OPS = 50000000
    DELEGATION_OPS = 1000000
    UNSTAKE_TOKENS = 5000000
    UNBOND_TOKENS = 5000000


def get_proxy() -> str:
    return get_value("proxy")


def get_chain_id() -> str:
    return get_value("chainID")


def get_tx_version() -> int:
    return int(get_value("txVersion"))


def get_dependency_tag(key: str) -> str:
    return get_value(f"dependencies.{key}.tag")


def set_dependency_tag(key: str, tag: str):
    set_value(f"dependencies.{key}.tag", tag)


def get_dependency_url(key: str, tag: str, platform: str) -> str:
    url_template = get_value(f"dependencies.{key}.urlTemplate.{platform}")
    return url_template.replace("{TAG}", tag)


def get_value(name: str) -> str:
    _guard_valid_name(name)
    data = get_active()
    return data.get(name, get_defaults()[name])


def set_value(name: str, value: Any):
    _guard_valid_name(name)
    data = read_file()
    active_config = data.get("active", "default")
    data.setdefault("configurations", {})
    data["configurations"].setdefault(active_config, {})
    data["configurations"][active_config][name] = value
    write_file(data)


def get_active():
    data = read_file()
    configs = data.get("configurations", {})
    active_config = data.get("active", "default")

    return configs.get(active_config, {})


def set_active(name: str):
    data = read_file()
    _guard_valid_config_name(data, name)
    data["active"] = name
    write_file(data)


def create_new_config(name: str, template: str):
    data = read_file()
    _guard_config_unique(data, name)
    new_config = {}
    if template is not None and template != "":
        _guard_valid_config_name(data, template)
        new_config = data["configurations"][template]

    data["active"] = name
    data.setdefault('configurations', {})
    data["configurations"][name] = new_config
    write_file(data)


def delete_config(name: str):
    _guard_valid_config_deletion(name)
    data = read_file()
    data["configurations"].pop(name, None)
    if data["active"] == name:
        data["active"] = "default"
    write_file(data)


def _guard_valid_name(name: str):
    if name not in get_defaults().keys():
        raise errors.UnknownConfigurationError(name)


def _guard_valid_config_name(config: Any, name: str):
    configurations = config.get('configurations', {})
    if name not in configurations:
        raise errors.UnknownConfigurationError(name)


def _guard_config_unique(config: Any, name: str):
    configurations = config.get('configurations', {})
    if name in configurations:
        raise errors.ConfigurationShouldBeUniqueError(name)


def _guard_valid_config_deletion(name: str):
    if name == "default":
        raise errors.ConfigurationProtectedError(name)


def get_defaults() -> Dict[str, Any]:
    return {
        "proxy": "https://testnet-gateway.elrond.com",
        "chainID": "T",
        "txVersion": "1",
        "dependencies.arwentools.tag": "v1.1.2",
        "dependencies.arwentools.urlTemplate.linux": "https://github.com/ElrondNetwork/arwen-wasm-vm/archive/{TAG}.tar.gz",
        "dependencies.arwentools.urlTemplate.osx": "https://github.com/ElrondNetwork/arwen-wasm-vm/archive/{TAG}.tar.gz",
        "dependencies.llvm.tag": "v9-19feb",
        "dependencies.llvm.urlTemplate.linux": "https://ide.elrond.com/vendor-llvm/{TAG}/linux-amd64.tar.gz?t=19feb",
        "dependencies.llvm.urlTemplate.osx": "https://ide.elrond.com/vendor-llvm/{TAG}/darwin-amd64.tar.gz?t=19feb",
        "dependencies.rust.tag": "",
        "dependencies.nodejs.tag": "v12.18.3",
        "dependencies.nodejs.urlTemplate.linux": "https://nodejs.org/dist/{TAG}/node-{TAG}-linux-x64.tar.gz",
        "dependencies.nodejs.urlTemplate.osx": "https://nodejs.org/dist/{TAG}/node-{TAG}-darwin-x64.tar.gz",
        "dependencies.elrond_go.tag": "master",
        "dependencies.elrond_go.urlTemplate.linux": "https://github.com/ElrondNetwork/elrond-go/archive/{TAG}.tar.gz",
        "dependencies.elrond_go.urlTemplate.osx": "https://github.com/ElrondNetwork/elrond-go/archive/{TAG}.tar.gz",
        "dependencies.elrond_go.url": "https://github.com/ElrondNetwork/elrond-go/archive/{TAG}.tar.gz",
        "dependencies.elrond_proxy_go.tag": "master",
        "dependencies.elrond_proxy_go.urlTemplate.linux": "https://github.com/ElrondNetwork/elrond-proxy-go/archive/{TAG}.tar.gz",
        "dependencies.elrond_proxy_go.urlTemplate.osx": "https://github.com/ElrondNetwork/elrond-proxy-go/archive/{TAG}.tar.gz",
        "dependencies.golang.tag": "go1.15.2",
        "dependencies.golang.urlTemplate.linux": "https://golang.org/dl/{TAG}.linux-amd64.tar.gz",
        "dependencies.golang.urlTemplate.osx": "https://golang.org/dl/{TAG}.darwin-amd64.tar.gz",
        "dependencies.mcl_signer.tag": "v1.0.0",
        "dependencies.mcl_signer.urlTemplate.linux": "https://github.com/ElrondNetwork/elrond-sdk-go-tools/releases/download/{TAG}/mcl_signer_{TAG}_ubuntu-latest.tar.gz",
        "dependencies.mcl_signer.urlTemplate.osx": "https://github.com/ElrondNetwork/elrond-sdk-go-tools/releases/download/{TAG}/mcl_signer_{TAG}_macos-latest.tar.gz",
        "dependencies.elrond_wasm_rs.tag": "v0.10.2",
    }


def read_file() -> Dict[str, Any]:
    if not os.path.isfile(CONFIG_PATH):
        return dict()
    return utils.read_json_file(CONFIG_PATH)


def write_file(data: Dict[str, Any]):
    utils.write_json_file(CONFIG_PATH, data)



from erdpy.accounts import Address
from erdpy.contracts import SmartContract
from erdpy.testnet import wallets


def get_owner_of_genesis_contracts():
    users = wallets.get_users()
    return users["alice"]


def get_delegation_address() -> Address:
    contract = SmartContract()
    contract.owner = get_owner_of_genesis_contracts()
    contract.owner.nonce = 0
    contract.compute_address()
    return contract.address


def is_last_user(nickname: str) -> bool:
    return nickname == "mike"


def is_foundational_node(nickname: str) -> bool:
    return nickname == "validator00"



class SmartContractViewModel:
    def __init__(self, project):
        self.FriendlyId = None  # this.FriendlyId = this.SourceFile.PathRelativeToWorkspace;
        self.SourceFile = MyFileViewModel()
        self.BytecodeFile = MyFileViewModel()
        self.PropertiesOnNodeDebug = PropertiesOnNetworkViewModel()
        self.PropertiesOnTestnet = PropertiesOnNetworkViewModel()
        self.IsBuilt = self.BytecodeFile is not None


class MyFileViewModel:
    def __init__(self):
        self.PathRelativeToWorkspace = None


class PropertiesOnNetworkViewModel:
    def __init__(self):
        self.Address = None
        self.AddressTimestamp = None
        self.LatestRun = SmartContractRunViewModel()
        self.WatchedVariables = []
        self.WatchedVariablesValues = dict()


class SmartContractRunViewModel:
    def __init__(self):
        self.Options = {
            "senderAddress": "",
            "functionName": "your_function",
            "functionArgs": [],
            "value": 0,
            "gasLimit": 500000000,
            "gasPrice": 200000000000000
        }

        self.VMOutput = dict()


class WatchedVariableViewModel:
    def __init__(self):
        self.Name = None
        self.FunctionName = None
        self.Arguments = None



import logging
import os
from binascii import unhexlify
from os import path
from typing import Any, Union

from erdpy import constants, errors, utils
from erdpy.interfaces import IAccount, IAddress
from erdpy.wallet import bech32, generate_pair, pem
from erdpy.wallet.keyfile import get_password, load_from_key_file

logger = logging.getLogger("accounts")


class AccountsRepository:
    def __init__(self, folder):
        utils.ensure_folder(folder)
        self.folder = folder

    def get_account(self, name):
        pem_file = path.join(self.folder, f"{name}.pem")
        return Account(pem_file=pem_file)

    def generate_accounts(self, count):
        for i in range(count):
            self.generate_account(i)

    def generate_account(self, name):
        seed, pubkey = generate_pair()
        address = Address(pubkey).bech32()

        pem_file = f"{name}_{address}.pem"
        pem_file = path.join(self.folder, pem_file)
        pem.write(pem_file, seed, pubkey, name=f"{name}:{address}")

    def get_all(self):
        accounts = []
        for pem_file in os.listdir(self.folder):
            pem_file = path.join(self.folder, pem_file)
            account = Account(pem_file=pem_file)
            accounts.append(account)

        return accounts


class Account(IAccount):
    def __init__(self, address: Any = None, pem_file: Union[str, None] = None, pem_index: int = 0, key_file: str = "", pass_file: str = ""):
        self.address = Address(address)
        self.pem_file = pem_file
        self.pem_index = int(pem_index)
        self.nonce: int = 0

        if pem_file:
            seed, pubkey = pem.parse(self.pem_file, self.pem_index)
            self.private_key_seed = seed.hex()
            self.address = Address(pubkey)
        elif key_file and pass_file:
            password = get_password(pass_file)
            address_from_key_file, seed = load_from_key_file(key_file, password)
            self.private_key_seed = seed.hex()
            self.address = Address(address_from_key_file)

    def sync_nonce(self, proxy: Any):
        logger.info("Account.sync_nonce()")
        self.nonce = proxy.get_account_nonce(self.address)
        logger.info(f"Account.sync_nonce() done: {self.nonce}")

    def get_seed(self) -> bytes:
        return unhexlify(self.private_key_seed)


class Address(IAddress):
    HRP = "erd"
    PUBKEY_LENGTH = 32
    PUBKEY_STRING_LENGTH = PUBKEY_LENGTH * 2    # hex-encoded
    BECH32_LENGTH = 62

    def __init__(self, value):
        self._value_hex = None

        if not value:
            return

        # Copy-constructor
        if isinstance(value, Address):
            value = value._value_hex

        # We keep a hex-encoded string as the "backing" value
        if len(value) == Address.PUBKEY_LENGTH:
            self._value_hex = value.hex()
        elif len(value) == Address.PUBKEY_STRING_LENGTH:
            self._value_hex = _as_string(value)
        elif len(value) == Address.BECH32_LENGTH:
            self._value_hex = _decode_bech32(value).hex()
        else:
            raise errors.BadAddressFormatError(value)

    def hex(self) -> str:
        self._assert_validity()
        return self._value_hex

    def bech32(self) -> str:
        self._assert_validity()
        pubkey = self.pubkey()
        return bech32.bech32_encode(self.HRP, bech32.convertbits(pubkey, 8, 5))

    def pubkey(self):
        self._assert_validity()
        pubkey = bytes.fromhex(self._value_hex)
        return pubkey

    def is_contract_address(self):
        return self.hex().startswith(constants.SC_HEX_PUBKEY_PREFIX)

    def _assert_validity(self):
        if self._value_hex is None:
            raise errors.EmptyAddressError()

    def __repr__(self):
        return self.bech32()

    @classmethod
    def zero(cls):
        return Address("0" * 64)


def _as_string(value):
    if isinstance(value, str):
        return value
    return value.decode("utf-8")


def _decode_bech32(value):
    bech32_string = _as_string(value)
    hrp, value_bytes = bech32.bech32_decode(bech32_string)
    if hrp != Address.HRP:
        raise errors.BadAddressFormatError(value)
    decoded_bytes = bech32.convertbits(value_bytes, 5, 8, False)
    return bytearray(decoded_bytes)



import binascii
import sys
from typing import Any

from erdpy import cli_shared, errors, utils
from erdpy.accounts import Address
from erdpy.delegation import staking_provider
from erdpy.proxy import ElrondProxy
from erdpy.transactions import do_prepare_transaction


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "staking-provider", "Staking provider omnitool")
    subparsers = parser.add_subparsers()

    # create new delegation contract
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "create-new-delegation-contract",
                                           "Create a new delegation system smart contract, transferred value must be"
                                           "greater than baseIssuingCost + min deposit value")
    _add_common_arguments(sub)
    sub.add_argument("--total-delegation-cap", required=True, help="the total delegation contract capacity")
    sub.add_argument("--service-fee", required=True, help="the delegation contract service fee")
    sub.set_defaults(func=do_create_delegation_contract)

    # get contract address
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "get-contract-address",
                                           "Get create contract address by transaction hash")
    sub.add_argument("--create-tx-hash", required=True, help="the hash")
    sub.add_argument("--sender", required=False, help="the sender address")
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_omit_fields_arg(sub)
    sub.set_defaults(func=get_contract_address_by_deploy_tx_hash)

    # add a new node
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "add-nodes",
                                           "Add new nodes must be called by the contract owner")
    sub.add_argument("--validators-file", required=True, help="a JSON file describing the Nodes")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    sub.add_argument("--using-delegation-manager", action="store_true", required=False, help="whether delegation contract was created using the Delegation Manager")
    _add_common_arguments(sub)
    sub.set_defaults(func=add_new_nodes)

    # remove nodes
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "remove-nodes",
                                           "Remove nodes must be called by the contract owner")
    sub.add_argument("--bls-keys", required=True, help="a list with the bls keys of the nodes")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=remove_nodes)

    # stake nodes
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "stake-nodes",
                                           "Stake nodes must be called by the contract owner")
    sub.add_argument("--bls-keys", required=True, help="a list with the bls keys of the nodes")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=stake_nodes)

    # unbond nodes
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "unbond-nodes",
                                           "Unbond nodes must be called by the contract owner")
    sub.add_argument("--bls-keys", required=True, help="a list with the bls keys of the nodes")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=unbond_nodes)

    # unstake nodes
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "unstake-nodes",
                                           "Unstake nodes must be called by the contract owner")
    sub.add_argument("--bls-keys", required=True, help="a list with the bls keys of the nodes")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=unstake_nodes)

    # unjail nodes
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "unjail-nodes",
                                           "Unjail nodes must be called by the contract owner")
    sub.add_argument("--bls-keys", required=True, help="a list with the bls keys of the nodes")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=unjail_nodes)

    # change service fee
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "change-service-fee",
                                           "Change service fee must be called by the contract owner")
    sub.add_argument("--service-fee", required=True, help="new service fee value")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=change_service_fee)

    # modify total delegation cap
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "modify-delegation-cap",
                                           "Modify delegation cap must be called by the contract owner")
    sub.add_argument("--delegation-cap", required=True, help="new delegation contract capacity")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=modify_delegation_cap)

    # set automatic activation
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "automatic-activation",
                                           "Automatic activation must be called by the contract owner")

    sub.add_argument("--set", action="store_true", required=not (utils.is_arg_present("--unset", sys.argv)),
                     help="set automatic activation True")
    sub.add_argument("--unset", action="store_true", required=not (utils.is_arg_present("--set", sys.argv)),
                     help="set automatic activation False")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=automatic_activation)

    # set redelegate cap
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "redelegate-cap",
                                           "Redelegate cap must be called by the contract owner")

    sub.add_argument("--set", action="store_true", required=not (utils.is_arg_present("--unset", sys.argv)),
                     help="set redelegate cap True")
    sub.add_argument("--unset", action="store_true", required=not (utils.is_arg_present("--set", sys.argv)),
                     help="set redelegate cap False")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=redelegate_cap)

    # set metadata
    sub = cli_shared.add_command_subparser(subparsers, "staking-provider", "set-metadata",
                                           "Set metadata must be called by the contract owner")

    sub.add_argument("--name", required=True, help="name field in staking provider metadata")
    sub.add_argument("--website", required=True, help="website field in staking provider metadata")
    sub.add_argument("--identifier", required=True, help="identifier field in staking provider metadata")
    sub.add_argument("--delegation-contract", required=True, help="address of the delegation contract")
    _add_common_arguments(sub)
    sub.set_defaults(func=set_metadata)


def _add_common_arguments(sub: Any):
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_wallet_args(sub)
    cli_shared.add_tx_args(sub, with_receiver=False, with_data=False, with_estimate_gas=True)
    cli_shared.add_broadcast_args(sub, relay=False)
    cli_shared.add_outfile_arg(sub, what="signed transaction, hash")


def do_create_delegation_contract(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_for_create_new_staking_contract(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def get_contract_address_by_deploy_tx_hash(args: Any):
    args = utils.as_object(args)
    omit_fields = cli_shared.parse_omit_fields_arg(args)

    proxy = ElrondProxy(args.proxy)

    transaction = proxy.get_transaction(args.create_tx_hash, with_results=True)
    utils.omit_fields(transaction, omit_fields)
    _get_sc_address_from_tx(transaction)


def add_new_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_for_add_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def remove_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_for_remove_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def stake_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_for_stake_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def unbond_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_for_unbond_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def unstake_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_for_unstake_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def unjail_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_for_unjail_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def change_service_fee(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_change_service_fee(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def modify_delegation_cap(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_modify_delegation_cap(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def automatic_activation(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_automatic_activation(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def redelegate_cap(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_redelegate_cap(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def set_metadata(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    staking_provider.prepare_args_set_metadata(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def _get_sc_address_from_tx(data: Any):
    if not isinstance(data, dict):
        raise errors.ProgrammingError("error")

    sc_results = data.get('smartContractResults')
    if sc_results is None:
        raise errors.ProgrammingError("smart contract results missing")

    # TODO improve robustness of this code in case of failed transaction
    try:
        sc_result = sc_results[0]
        data_field = sc_result['data']

        data_field_split = data_field.split('@')
        arg_1 = binascii.unhexlify(data_field_split[1])
        if not arg_1 == b'ok':
            raise errors.ProgrammingError(arg_1.decode("utf-8"))

        sc_address = binascii.unhexlify(data_field_split[2])
        address = Address(sc_address)
        print("Contract address: ", address)
    except Exception:
        raise errors.ProgrammingError(
            "cannot get the smart contract address from transaction results, please try again")



import argparse
import ast
import sys
from argparse import FileType
from typing import Any, List, Text

from erdpy import config, errors, scope, utils
from erdpy.accounts import Account
from erdpy.proxy.core import ElrondProxy
from erdpy.transactions import Transaction


def wider_help_formatter(prog: Text):
    return argparse.HelpFormatter(prog, max_help_position=50, width=120)


def add_group_subparser(subparsers: Any, group: str, description: str) -> Any:
    parser = subparsers.add_parser(
        group,
        usage=f"erdpy {group} COMMAND [-h] ...",
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser._positionals.title = "COMMANDS"
    parser._optionals.title = "OPTIONS"

    return parser


def build_group_epilog(subparsers: Any) -> str:
    epilog = """
----------------
COMMANDS summary
----------------
"""
    for choice, sub in subparsers.choices.items():
        epilog += f"{choice.ljust(30)} {sub.description}\n"

    return epilog


def add_command_subparser(subparsers: Any, group: str, command: str, description: str):
    return subparsers.add_parser(
        command,
        usage=f"erdpy {group} {command} [-h] ...",
        description=description,
        formatter_class=wider_help_formatter
    )


def add_tx_args(sub: Any, with_nonce: bool = True, with_receiver: bool = True, with_data: bool = True, with_estimate_gas: bool = False):
    if with_nonce:
        sub.add_argument("--nonce", type=int, required=not("--recall-nonce" in sys.argv), help="# the nonce for the transaction")
        sub.add_argument("--recall-nonce", action="store_true", default=False, help="⭮ whether to recall the nonce when creating the transaction (default: %(default)s)")

    if with_receiver:
        sub.add_argument("--receiver", required=True, help="🖄 the address of the receiver")
        sub.add_argument("--receiver-username", required=False, help="🖄 the username of the receiver")

    sub.add_argument("--gas-price", default=config.DEFAULT_GAS_PRICE, help="⛽ the gas price (default: %(default)d)")
    sub.add_argument("--gas-limit", required=not("--estimate-gas" in sys.argv), help="⛽ the gas limit")
    if with_estimate_gas:
        sub.add_argument("--estimate-gas", action="store_true", default=False, help="⛽ whether to estimate the gas limit (default: %(default)d)")

    sub.add_argument("--value", default="0", help="the value to transfer (default: %(default)s)")

    if with_data:
        sub.add_argument("--data", default="", help="the payload, or 'memo' of the transaction (default: %(default)s)")

    sub.add_argument("--chain", default=scope.get_chain_id(), help="the chain identifier (default: %(default)s)")
    sub.add_argument("--version", type=int, default=scope.get_tx_version(), help="the transaction version (default: %(default)s)")


def add_wallet_args(sub: Any):
    sub.add_argument("--pem", required=not (utils.is_arg_present("--keyfile", sys.argv)), help="🔑 the PEM file, if keyfile not provided")
    sub.add_argument("--pem-index", default=0, help="🔑 the index in the PEM file (default: %(default)s)")
    sub.add_argument("--keyfile", required=not (utils.is_arg_present("--pem", sys.argv)), help="🔑 a JSON keyfile, if PEM not provided")
    sub.add_argument("--passfile", required=not (utils.is_arg_present("--pem", sys.argv)), help="🔑 a file containing keyfile's password, if keyfile provided")
    sub.add_argument("--sender-username", required=False, help="🖄 the username of the sender")


def add_proxy_arg(sub: Any):
    sub.add_argument("--proxy", default=scope.get_proxy(), help="🖧 the URL of the proxy (default: %(default)s)")


def add_outfile_arg(sub: Any, what: str = ""):
    what = f"({what})" if what else ""
    sub.add_argument("--outfile", type=FileType("w"), default=sys.stdout, help=f"where to save the output {what} (default: stdout)")


def add_infile_arg(sub: Any, what: str = ""):
    what = f"({what})" if what else ""
    sub.add_argument("--infile", type=FileType("r"), required=True, help=f"input file {what}")


def add_omit_fields_arg(sub: Any):
    sub.add_argument("--omit-fields", default="[]", type=str, required=False, help="omit fields in the output payload (default: %(default)s)")


def parse_omit_fields_arg(args: Any) -> List[str]:
    literal = args.omit_fields
    parsed = ast.literal_eval(literal)
    return parsed


def prepare_nonce_in_args(args: Any):
    if args.recall_nonce:
        if args.pem:
            account = Account(pem_file=args.pem, pem_index=args.pem_index)
        elif args.keyfile and args.passfile:
            account = Account(key_file=args.keyfile, pass_file=args.passfile)
        else:
            raise errors.NoWalletProvided()

        account.sync_nonce(ElrondProxy(args.proxy))
        args.nonce = account.nonce


def add_broadcast_args(sub: Any, simulate=True, relay=False):
    sub.add_argument("--send", action="store_true", default=False, help="✓ whether to broadcast the transaction (default: %(default)s)")

    if simulate:
        sub.add_argument("--simulate", action="store_true", default=False, help="whether to simulate the transaction (default: %(default)s)")
    if relay:
        sub.add_argument("--relay", action="store_true", default=False, help="whether to relay the transaction (default: %(default)s)")


def check_broadcast_args(args: Any):
    if hasattr(args, "relay") and args.relay and args.send:
        raise errors.BadUsage("Cannot directly send a relayed transaction. Use 'erdpy tx new --relay' first, then 'erdpy tx send --data-file'")
    if args.send and args.simulate:
        raise errors.BadUsage("Cannot both 'simulate' and 'send' a transaction")


def send_or_simulate(tx: Transaction, args: Any):
    if args.send:
        tx.send(ElrondProxy(args.proxy))
    elif args.simulate:
        response = tx.simulate(ElrondProxy(args.proxy))
        utils.dump_out_json(response)



import logging
import os
from pathlib import Path
from typing import Any, Dict

from erdpy import cli_shared, errors, utils, workstation

logger = logging.getLogger("cli.data")

DATA_FILENAME = "erdpy.data-storage.json"


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "data", "Data manipulation omnitool")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "data", "parse", "Parses values from a given file")
    sub.add_argument("--file", required=True, help="path of the file to parse")
    sub.add_argument("--expression", required=True, help="the Python-Dictionary expression to evaluate in order to extract the data")
    sub.set_defaults(func=parse)

    sub = cli_shared.add_command_subparser(subparsers, "data", "store", "Stores a key-value pair within a partition")
    sub.add_argument("--key", required=True, help="the key")
    sub.add_argument("--value", required=True, help="the value to save")
    sub.add_argument("--partition", default="*", help="the storage partition (default: %(default)s)")
    sub.add_argument("--use-global", action="store_true", default=False, help="use the global storage (default: %(default)s)")
    sub.set_defaults(func=store)

    sub = cli_shared.add_command_subparser(subparsers, "data", "load", "Loads a key-value pair from a storage partition")
    sub.add_argument("--key", required=True, help="the key")
    sub.add_argument("--partition", default="*", help="the storage partition (default: %(default)s)")
    sub.add_argument("--use-global", action="store_true", default=False, help="use the global storage (default: %(default)s)")
    sub.set_defaults(func=load)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def parse(args: Any):
    logger.warning("Always review --expression parameters before executing this command!")

    file = Path(args.file).expanduser()
    expression: str = args.expression
    suffix = file.suffix
    data = None

    if suffix == ".json":
        data = utils.read_json_file(str(file))
    else:
        raise errors.BadUsage(f"File isn't parsable: {file}")

    try:
        result = eval(expression, {
            "data": data
        })
    except KeyError:
        result = ""

    print(result)


def store(args: Any):
    logger.warning("Never use this command to store sensitive information! Data is unencrypted.")

    key = args.key
    value = args.value
    partition = args.partition
    use_global = args.use_global

    data = _read_file(use_global)
    if partition not in data:
        data[partition] = dict()

    data_in_partition = data[partition]
    data_in_partition[key] = value
    _write_file(use_global, data)

    logger.info(f"Data has been stored at key = '{key}', in partition = '{partition}'.")


def load(args: Any):
    key = args.key
    partition = args.partition
    use_global = args.use_global

    data = _read_file(use_global)
    data_in_partition = data.get(partition, dict())
    value = data_in_partition.get(key, "")
    print(value)


def _read_file(use_global: bool) -> Dict[str, Any]:
    filename = _get_filename(use_global)

    if not os.path.isfile(filename):
        return dict()
    return utils.read_json_file(filename)


def _write_file(use_global: bool, data: Dict[str, Any]):
    filename = _get_filename(use_global)

    utils.write_json_file(filename, data)


def _get_filename(use_global: bool):
    if use_global:
        return workstation.get_tools_folder() / DATA_FILENAME
    return Path(os.getcwd()) / DATA_FILENAME



from typing import Any

from erdpy import cli_shared, utils
from erdpy.proxy.core import ElrondProxy


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "hyperblock", "Get Hyperblock from the Network")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "hyperblock", "get", "Get hyperblock")

    cli_shared.add_proxy_arg(sub)
    sub.add_argument("--key", required=True, help="the hash or the nonce of the hyperblock")

    sub.set_defaults(func=get_hyperblock)


def get_hyperblock(args: Any) -> Any:
    proxy_url = args.proxy
    proxy = ElrondProxy(proxy_url)
    response = proxy.get_hyperblock(args.key)
    utils.dump_out_json(response)



import logging
from typing import Any

from erdpy import cli_shared, config, dependencies, errors
from erdpy.dependencies.install import get_deps_dict

logger = logging.getLogger("cli.deps")


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "deps", "Manage dependencies or elrond-sdk modules")
    subparsers = parser.add_subparsers()

    choices = list(get_deps_dict().keys())

    sub = cli_shared.add_command_subparser(subparsers, "deps", "install", "Install dependencies or elrond-sdk modules.")
    sub.add_argument("name", choices=choices, help="the dependency to install")
    sub.add_argument("--overwrite", action="store_true", default=False, help="whether to overwrite an existing installation")
    sub.add_argument("--tag", help="the tag or version to install")
    sub.set_defaults(func=install)

    sub = cli_shared.add_command_subparser(subparsers, "deps", "check", "Check whether a dependency is installed.")
    sub.add_argument("name", choices=choices, help="the dependency to check")
    sub.add_argument("--tag", help="the tag or version to check")
    sub.set_defaults(func=check)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def install(args: Any):
    name: str = args.name
    tag: str = args.tag
    overwrite: bool = args.overwrite
    dependencies.install_module(name, tag, overwrite)


def check(args: Any):
    name: str = args.name
    tag: str = args.tag
    module = dependencies.get_module_by_key(name)
    default_tag: str = config.get_dependency_tag(module.key)
    tag_to_check = tag or default_tag
    installed = module.is_installed(tag_to_check)
    if installed:
        print(f"[{name} {tag_to_check}] is installed. Default version (tag) is [{default_tag}].")
        return

    raise errors.DependencyMissing(name, tag_to_check)



import logging
from typing import Dict, List

from erdpy import config, errors
from erdpy.dependencies.modules import (ArwenToolsModule, DependencyModule,
                                        GolangModule, MclSignerModule,
                                        NodejsModule, Rust, StandaloneModule)

logger = logging.getLogger("install")


def install_module(key: str, tag: str = "", overwrite: bool = False):
    module = get_module_by_key(key)
    module.install(tag, overwrite)


def get_module_directory(key: str) -> str:
    module = get_module_by_key(key)
    default_tag = config.get_dependency_tag(key)
    directory = module.get_directory(default_tag)
    return directory


def get_module_by_key(key: str) -> DependencyModule:
    matches = [module for module in get_all_deps() if module.key == key or key in module.aliases]
    if len(matches) != 1:
        raise errors.UnknownDependency(key)

    return matches[0]


def get_deps_dict() -> Dict[str, DependencyModule]:
    deps = dict()
    for module in get_all_deps():
        deps[module.key] = module
        for alias in module.aliases:
            deps[alias] = module
    return deps


def get_all_deps() -> List[DependencyModule]:
    return [
        StandaloneModule(key="llvm", aliases=["clang", "cpp"]),
        ArwenToolsModule(key="arwentools"),
        Rust(key="rust"),
        NodejsModule(key="nodejs", aliases=[]),
        StandaloneModule(key="elrond_go"),
        StandaloneModule(key="elrond_proxy_go"),
        GolangModule(key="golang"),
        MclSignerModule(key="mcl_signer")
    ]



import binascii
import logging
from os import path
from typing import Any

from erdpy.validators.validators_file import ValidatorsFile
from erdpy.conv.conv import Converters
from erdpy.accounts import Account, Address
from erdpy.config import MetaChainSystemSCsCost, MIN_GAS_LIMIT, GAS_PER_DATA_BYTE
from erdpy.wallet.pem import parse_validator_pem
from erdpy.wallet.signing import sign_message_with_bls_key

logger = logging.getLogger("validators")

VALIDATORS_SMART_CONTRACT_ADDRESS = "erd1qqqqqqqqqqqqqqqpqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqplllst77y4l"


def prepare_args_for_stake(args: Any):
    if args.top_up:
        prepare_args_for_top_up(args)
        return

    validators_file = ValidatorsFile(args.validators_file)

    reward_address = args.reward_address

    # TODO: Refactor, so that only address is received here.
    if args.pem:
        account = Account(pem_file=args.pem)
    elif args.keyfile and args.passfile:
        account = Account(key_file=args.keyfile, pass_file=args.passfile)

    num_of_nodes = validators_file.get_num_of_nodes()
    stake_data = 'stake@' + binascii.hexlify(num_of_nodes.to_bytes(1, byteorder="little")).decode()
    validators_list = validators_file.get_validators_list()
    for validator in validators_list:
        # get validator
        validator_pem = validator.get("pemFile")
        validator_pem = path.join(path.dirname(args.validators_file), validator_pem)
        seed, bls_key = parse_validator_pem(validator_pem)
        signed_message = sign_message_with_bls_key(account.address.pubkey().hex(), seed.decode('ascii'))
        stake_data += f"@{bls_key}@{signed_message}"

    if reward_address:
        reward_address = Address(args.reward_address)
        stake_data += '@' + reward_address.hex()

    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS
    args.data = stake_data

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.STAKE, num_of_nodes)


def prepare_args_for_top_up(args: Any):
    args.data = 'stake'
    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.STAKE, 1)


def prepare_args_for_unstake(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.nodes_public_keys)
    args.data = 'unStake' + parsed_keys
    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNSTAKE, num_keys)


def prepare_args_for_unbond(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.nodes_public_keys)
    args.data = 'unBond' + parsed_keys
    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNBOND, num_keys)


def prepare_args_for_unjail(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.nodes_public_keys)
    args.data = 'unJail' + parsed_keys
    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNJAIL, num_keys)


def prepare_args_for_change_reward_address(args: Any):
    reward_address = Address(args.reward_address)
    args.data = 'changeRewardAddress@' + reward_address.hex()
    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.CHANGE_REWARD_ADDRESS)


def prepare_args_for_claim(args: Any):
    args.data = 'claim'
    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS

    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.CLAIM)


def prepare_args_for_unstake_nodes(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.nodes_public_keys)
    args.data = 'unStakeNodes' + parsed_keys

    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNSTAKE, num_keys)


def prepare_args_for_unstake_tokens(args: Any):
    args.data = 'unStakeTokens'
    args.data += '@' + Converters.str_int_to_hex_str(str(args.unstake_value))

    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNSTAKE_TOKENS)


def prepare_args_for_unbond_nodes(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.nodes_public_keys)
    args.data = 'unBondNodes' + parsed_keys

    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNBOND, num_keys)


def prepare_args_for_unbond_tokens(args: Any):
    args.data = 'unBondTokens'
    args.data += '@' + Converters.str_int_to_hex_str(str(args.unbond_value))

    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.UNBOND_TOKENS)


def prepare_args_for_clean_registered_data(args: Any):
    args.data = 'cleanRegisteredData'

    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.STAKE)


def prepare_args_for_restake_unstaked_nodes(args: Any):
    parsed_keys, num_keys = Converters.parse_keys(args.nodes_public_keys)
    args.data = 'reStakeUnStakedNodes' + parsed_keys

    args.receiver = VALIDATORS_SMART_CONTRACT_ADDRESS
    if args.estimate_gas:
        args.gas_limit = estimate_system_sc_call(args, MetaChainSystemSCsCost.STAKE, num_keys)


def estimate_system_sc_call(args, base_cost, factor=1):
    num_bytes = len(args.data)
    gas_limit = MIN_GAS_LIMIT + num_bytes * GAS_PER_DATA_BYTE
    gas_limit += factor * base_cost
    return gas_limit



import hashlib
import hmac
import struct

import nacl.signing

BIP39_SALT_MODIFIER = "mnemonic"
BIP39_PBKDF2_ROUNDS = 2048
BIP32_SEED_MODIFIER = b'ed25519 seed'
ELROND_DERIVATION_PATH = [44, 508, 0, 0]
HARDENED_OFFSET = 0x80000000


def derive_keys(mnemonic, account_index=0):
    seed = mnemonic_to_bip39seed(mnemonic)
    private_key = bip39seed_to_private_key(seed, account_index)
    public_key = bytes(nacl.signing.SigningKey(private_key).verify_key)
    return private_key, public_key


# References:
# https://github.com/trezor/python-mnemonic/blob/master/mnemonic/mnemonic.py
# https://ethereum.stackexchange.com/a/72871/59887
def mnemonic_to_bip39seed(mnemonic, passphrase=""):
    passphrase = BIP39_SALT_MODIFIER + passphrase
    mnemonic = mnemonic.encode("utf-8")
    passphrase = passphrase.encode("utf-8")
    stretched = hashlib.pbkdf2_hmac("sha512", mnemonic, passphrase, BIP39_PBKDF2_ROUNDS)
    return stretched[:64]


# References:
# https://ethereum.stackexchange.com/a/72871/59887s
# https://github.com/alepop/ed25519-hd-key/blob/master/src/index.ts#L22
def bip39seed_to_master_key(seed):
    hashed = hmac.new(BIP32_SEED_MODIFIER, seed, hashlib.sha512).digest()
    key, chain_code = hashed[:32], hashed[32:]
    return key, chain_code


# Reference: https://github.com/alepop/ed25519-hd-key
def bip39seed_to_private_key(seed, account_index=0):
    key, chain_code = bip39seed_to_master_key(seed)

    for segment in ELROND_DERIVATION_PATH + [account_index]:
        key, chain_code = _ckd_priv(key, chain_code, segment + HARDENED_OFFSET)

    return key


# Reference: https://github.com/alepop/ed25519-hd-key
def _ckd_priv(key, chain_code, index):
    index_buffer = struct.pack('>I', index)
    data = bytearray([0]) + bytearray(key) + bytearray(index_buffer)
    hashed = hmac.new(chain_code, data, hashlib.sha512).digest()
    key, chain_code = hashed[:32], hashed[32:]

    return key, chain_code



import sys
from pathlib import Path

from erdpy import config, utils


def get_tools_folder():
    folder = Path.home() / config.ROOT_FOLDER_NAME
    utils.ensure_folder(folder)
    return folder


def get_platform():
    platforms = {
        "linux": "linux",
        "linux1": "linux",
        "linux2": "linux",
        "darwin": "osx",
        "win32": "windows",
        "cygwin": "windows",
        "msys": "windows"
    }

    platform = platforms.get(sys.platform)
    if platform is None:
        raise Exception(f"Unknown platform: {sys.platform}")

    return platform



from typing import Any, List

from Cryptodome.Hash import keccak

from erdpy import cli_shared, utils
from erdpy.accounts import Account, Address
from erdpy.contracts import SmartContract
from erdpy.proxy.core import ElrondProxy
from erdpy.transactions import do_prepare_transaction

MaxNumShards = 256
ShardIdentiferLen = 2
InitialDNSAddress = bytes([1] * 32)


def resolve(name: str, proxy: ElrondProxy) -> Address:
    name_arg = "0x{}".format(str.encode(name).hex())
    dns_address = dns_address_for_name(name)
    contract = SmartContract(dns_address)
    result = contract.query(proxy, "resolve", [name_arg])
    if len(result) == 0:
        return Address.zero()
    return Address(result[0].hex)


def validate_name(name: str, shard_id: int, proxy: ElrondProxy):
    name_arg = "0x{}".format(str.encode(name).hex())
    dns_address = compute_dns_address_for_shard_id(shard_id)
    contract = SmartContract(dns_address)
    contract.query(proxy, "validateName", [name_arg])


def register(args: Any):
    args = utils.as_object(args)

    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    args.receiver = dns_address_for_name(args.name).bech32()
    args.data = dns_register_data(args.name)

    tx = do_prepare_transaction(args)

    if hasattr(args, "relay") and args.relay:
        args.outfile.write(tx.serialize_as_inner())
        return

    try:
        if args.send:
            tx.send(ElrondProxy(args.proxy))
        elif args.simulate:
            response = tx.simulate(ElrondProxy(args.proxy))
            utils.dump_out_json(response)
    finally:
        tx.dump_to(args.outfile)


def compute_all_dns_addresses() -> List[Address]:
    addresses = []
    for i in range(0, 256):
        addresses.append(compute_dns_address_for_shard_id(i))
    return addresses


def name_hash(name: str) -> Address:
    return keccak.new(digest_bits=256).update(str.encode(name)).digest()


def registration_cost(shard_id: int, proxy: ElrondProxy) -> int:
    dns_address = compute_dns_address_for_shard_id(shard_id)
    contract = SmartContract(dns_address)
    result = contract.query(proxy, "getRegistrationCost", [])
    if len(result[0]) == 0:
        return 0
    else:
        return int("0x{}".format(result[0]))


def version(shard_id: int, proxy: ElrondProxy) -> str:
    dns_address = compute_dns_address_for_shard_id(shard_id)
    contract = SmartContract(dns_address)
    result = contract.query(proxy, "version", [])
    return bytearray.fromhex(result[0].hex).decode()


def dns_address_for_name(name: str) -> Address:
    hash = name_hash(name)
    shard_id = hash[31]
    return compute_dns_address_for_shard_id(shard_id)


def compute_dns_address_for_shard_id(shard_id) -> Address:
    deployer_pubkey_prefix = InitialDNSAddress[:len(InitialDNSAddress) - ShardIdentiferLen]

    deployer_pubkey = deployer_pubkey_prefix + bytes([0, shard_id])
    deployer = Account(address=Address(deployer_pubkey))
    deployer.nonce = 0
    # Workaround: currently, in erdpy, in order to compute the address of a contract, one has to create an instance of the class "SmartContract".
    # This might change in the future.
    contract = SmartContract()
    contract.owner = deployer
    contract.compute_address()
    return contract.address


def dns_register_data(name: str) -> str:
    name_enc: bytes = str.encode(name)
    return "register@{}".format(name_enc.hex())



from typing import Any
from prettytable import PrettyTable

from erdpy import cli_shared
from erdpy.dns import name_hash, dns_address_for_name, register, resolve, registration_cost, validate_name, version, compute_dns_address_for_shard_id
from erdpy.accounts import Address
from erdpy.proxy.core import ElrondProxy


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "dns", "Operations related to the Domain Name Service")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "dns", "register", "Send a register transaction to the appropriate DNS contract from given user and with given name")
    cli_shared.add_outfile_arg(sub)
    cli_shared.add_broadcast_args(sub, relay=True)
    cli_shared.add_wallet_args(sub)
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_tx_args(sub, with_receiver=False, with_data=False)
    sub.add_argument("--name", help="the name to register")
    sub.set_defaults(func=register)

    sub = cli_shared.add_command_subparser(subparsers, "dns", "resolve", "Find the address for a name")
    _add_name_arg(sub)
    cli_shared.add_proxy_arg(sub)
    sub.set_defaults(func=dns_resolve)

    sub = cli_shared.add_command_subparser(subparsers, "dns", "validate-name", "Asks one of the DNS contracts to validate a name. Can be useful before registering it.")
    _add_name_arg(sub)
    sub.add_argument("--shard-id", default=0, help="shard id of the contract to call (default: %(default)s)")
    cli_shared.add_proxy_arg(sub)
    sub.set_defaults(func=dns_validate_name)

    sub = cli_shared.add_command_subparser(subparsers, "dns", "name-hash", "The hash of a name, as computed by a DNS smart contract")
    _add_name_arg(sub)
    sub.set_defaults(func=get_name_hash)

    sub = cli_shared.add_command_subparser(subparsers, "dns", "registration-cost", "Gets the registration cost from a DNS smart contract, by default the one with shard id 0.")
    sub.add_argument("--shard-id", default=0, help="shard id of the contract to call (default: %(default)s)")
    cli_shared.add_proxy_arg(sub)
    sub.set_defaults(func=get_registration_cost)

    sub = cli_shared.add_command_subparser(subparsers, "dns", "version", "Asks the contract for its version")
    sub.add_argument("--shard-id", default=0, help="shard id of the contract to call (default: %(default)s)")
    sub.add_argument("--all", action="store_true", default=False, help="prints a list of all DNS contracts and their current versions (default: %(default)s)")
    cli_shared.add_proxy_arg(sub)
    sub.set_defaults(func=get_version)

    sub = cli_shared.add_command_subparser(subparsers, "dns", "dns-addresses", "Lists all 256 DNS contract addresses")
    sub.set_defaults(func=print_dns_addresses_table)

    sub = cli_shared.add_command_subparser(subparsers, "dns", "dns-address-for-name", "DNS contract address (bech32) that corresponds to a name")
    _add_name_arg(sub)
    sub.set_defaults(func=get_dns_address_for_name)

    sub = cli_shared.add_command_subparser(subparsers, "dns", "dns-address-for-name-hex", "DNS contract address (hex) that corresponds to a name")
    _add_name_arg(sub)
    sub.set_defaults(func=get_dns_address_for_name_hex)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def _add_name_arg(sub: Any):
    sub.add_argument("name", help="the name for which to check")


def dns_resolve(args: Any):
    addr = resolve(args.name, ElrondProxy(args.proxy))
    if addr.hex() != Address.zero().hex():
        print(addr.bech32())


def dns_validate_name(args: Any):
    validate_name(args.name, args.shard_id, ElrondProxy(args.proxy))


def get_name_hash(args: Any):
    print(name_hash(args.name).hex())


def get_dns_address_for_name(args: Any):
    name = args.name
    dns_address = dns_address_for_name(name)
    print(dns_address)


def get_dns_address_for_name_hex(args: Any):
    name = args.name
    dns_address = dns_address_for_name(name)
    print(dns_address.hex())


def get_registration_cost(args: Any):
    print(registration_cost(args.shard_id, ElrondProxy(args.proxy)))


def get_version(args: Any):
    proxy = ElrondProxy(args.proxy)
    if args.all:
        t = PrettyTable(['Shard ID', 'Contract address (bech32)', 'Contract address (hex)', 'Version'])
        for shard_id in range(0, 256):
            address = compute_dns_address_for_shard_id(shard_id)
            v = version(shard_id, proxy)
            t.add_row([shard_id, address, address.hex(), v])
        print(t)
    else:
        shard_id = int(args.shard_id)
        print(version(shard_id, proxy))


def print_dns_addresses_table(args: Any):
    t = PrettyTable(['Shard ID', 'Contract address (bech32)', 'Contract address (hex)'])
    for shard_id in range(0, 256):
        address = compute_dns_address_for_shard_id(shard_id)
        t.add_row([shard_id, address, address.hex()])
    print(t)



import sys
from typing import Any

from erdpy import cli_shared, validators, utils
from erdpy.transactions import do_prepare_transaction


def setup_parser(subparsers: Any) -> Any:
    parser = cli_shared.add_group_subparser(subparsers, "validator", "Stake, UnStake, UnBond, Unjail and other "
                                                                     "actions useful for "
                                                                     "Validators")
    subparsers = parser.add_subparsers()

    sub = cli_shared.add_command_subparser(subparsers, "validator", "stake", "Stake value into the Network")
    _add_common_arguments(sub)
    sub.add_argument("--reward-address", default="", help="the reward address")
    sub.add_argument("--validators-file", required=not (utils.is_arg_present("--top-up", sys.argv)),
                     help="a JSON file describing the Nodes")
    sub.add_argument("--top-up", action="store_true", default=False,
                     required=not (utils.is_arg_present("--validators-file", sys.argv)), help="Stake value for top up")
    sub.set_defaults(func=do_stake)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unstake", "Unstake value")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_unstake)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unjail", "Unjail a Validator Node")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_unjail)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unbond", "Unbond tokens for a bls key")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_unbond)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "change-reward-address",
                                           "Change the reward address")
    _add_common_arguments(sub)
    sub.add_argument("--reward-address", required=True, help="the new reward address")
    sub.set_defaults(func=change_reward_address)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "claim", "Claim rewards")
    _add_common_arguments(sub)
    sub.set_defaults(func=do_claim)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unstake-nodes", "Unstake-nodes will unstake "
                                                                                     "nodes for provided bls keys")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_unstake_nodes)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unstake-tokens", "This command will un-stake the "
                                                                                      "given amount (if value is "
                                                                                      "greater than the existing "
                                                                                      "topUp value, it will unStake "
                                                                                      "one or several nodes)")
    _add_common_arguments(sub)
    sub.add_argument("--unstake-value", default=0, help="the unstake value")
    sub.set_defaults(func=do_unstake_tokens)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unbond-nodes", "It will unBond nodes")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_unbond_nodes)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "unbond-tokens", "It will unBond tokens, if "
                                                                                     "provided value is bigger that "
                                                                                     "topUp value will unBond nodes")
    _add_common_arguments(sub)
    sub.add_argument("--unbond-value", default=0, help="the unbond value")
    sub.set_defaults(func=do_unbond_tokens)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "clean-registered-data", "Deletes duplicated keys "
                                                                                             "from registered data")
    _add_common_arguments(sub)
    sub.set_defaults(func=do_clean_registered_data)

    sub = cli_shared.add_command_subparser(subparsers, "validator", "restake-unstaked-nodes", "It will reStake UnStaked nodes")
    _add_common_arguments(sub)
    _add_nodes_arg(sub)
    sub.set_defaults(func=do_restake_unstaked_nodes)

    parser.epilog = cli_shared.build_group_epilog(subparsers)
    return subparsers


def _add_common_arguments(sub: Any):
    cli_shared.add_proxy_arg(sub)
    cli_shared.add_wallet_args(sub)
    cli_shared.add_tx_args(sub, with_receiver=False, with_data=False, with_estimate_gas=True)
    cli_shared.add_broadcast_args(sub, relay=False)
    cli_shared.add_outfile_arg(sub, what="signed transaction, hash")


def _add_nodes_arg(sub: Any):
    sub.add_argument("--nodes-public-keys", required=True, help="the public keys of the nodes as CSV (addrA,addrB)")


def do_stake(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_stake(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_unstake(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_unstake(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_unjail(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_unjail(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_unbond(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_unbond(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def change_reward_address(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_change_reward_address(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_claim(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_claim(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_unstake_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_unstake_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_unstake_tokens(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_unstake_tokens(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_unbond_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_unbond_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_unbond_tokens(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_unbond_tokens(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_clean_registered_data(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_clean_registered_data(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)


def do_restake_unstaked_nodes(args: Any):
    cli_shared.check_broadcast_args(args)
    cli_shared.prepare_nonce_in_args(args)
    validators.prepare_args_for_restake_unstaked_nodes(args)
    tx = do_prepare_transaction(args)

    try:
        cli_shared.send_or_simulate(tx, args)
    finally:
        tx.dump_to(args.outfile)



import os.path
from erdpy import errors


def is_file(input):
    if not os.path.isfile(input):
        raise errors.BadInputError(input, "is not a valid file")


def is_directory(directory: str):
    if not os.path.isdir(directory):
        raise errors.BadDirectory(directory)


def is_hex_address(input):
    is_hex_string(input)

    if len(input) != 64:
        raise errors.BadInputError(input, "is not a valid hex-encoded address")


def is_hex_string(input):
    try:
        bytearray.fromhex(input)
    except Exception:
        raise errors.BadInputError(input, "is not a valid hex-encoded string")



import json
from os import path

from erdpy import guards
from erdpy.errors import CannotReadValidatorsData


class ValidatorsFile:
    def __init__(self, validators_file_path: str):
        self.validators_file_path = validators_file_path
        self._validators_data = self._read_json_file_validators()

    def get_num_of_nodes(self) -> int:
        return len(self._validators_data.get("validators", []))

    def get_validators_list(self):
        return self._validators_data.get("validators", [])

    def _read_json_file_validators(self):
        val_file = path.expanduser(self.validators_file_path)
        guards.is_file(val_file)
        with open(self.validators_file_path, "r") as json_file:
            try:
                data = json.load(json_file)
            except Exception:
                raise CannotReadValidatorsData()
            return data



import logging
import os
import subprocess
from os import path
from pathlib import Path

from erdpy import dependencies, errors, myprocess, utils
from erdpy.projects.project_base import Project

logger = logging.getLogger('ProjectClang')


class ProjectClang(Project):

    def __init__(self, directory):
        super().__init__(directory)
        self.ensure_config_file()

    def perform_build(self):
        self.config = self.load_config()
        self.ensure_source_files()

        self.unit = self.get_unit_file()
        self.file_ll = self.unit.with_suffix('.ll')
        self.file_o = self.unit.with_suffix('.o')
        self.file_export = self.unit.with_suffix('.export')
        self.file_output = self.unit.with_suffix('.wasm')

        try:
            self.do_clang()
            self.do_llvm_link()
            self.do_llc()
            self.do_wasm()
        except subprocess.CalledProcessError as err:
            raise errors.BuildError(err.output)

    def do_clang(self):
        logger.info('do_clang')

        tool = path.join(self._get_llvm_path(), 'clang-9')
        args = [
            tool,
            '-cc1', '-emit-llvm',
            '-triple=wasm32-unknown-unknown-wasm',
        ]

        if self.options.get('optimized', False):
            args.append('-Ofast')
        else:
            args.append('-O0')

        args.extend(map(str, self.get_source_files()))
        myprocess.run_process(args)

    def do_llvm_link(self):
        logger.info('do_llvm_link')
        tool = path.join(self._get_llvm_path(), 'llvm-link')
        args = [tool]
        args.extend(['-o', str(self.file_ll)])
        args.extend(map(str, self.get_ll_files()))
        myprocess.run_process(args)

    def do_llc(self):
        logger.info('do_llc')
        tool = path.join(self._get_llvm_path(), 'llc')
        args = [tool]

        if self.options.get('optimized', False):
            args.append('-O3')
        else:
            args.append('-O0')

        args.append('-filetype=obj')
        args.append(str(self.file_ll))

        args.extend(['-o', str(self.file_o)])
        myprocess.run_process(args)

    def do_wasm(self):
        logger.info('do_wasm')
        tool = path.join(self._get_llvm_path(), 'wasm-ld')
        args = [
            tool,
            '--no-entry',
            str(self.file_o),
            '-o', self.file_output,
            '--strip-all',
            '-allow-undefined'
        ]

        if self.options.get('verbose', False):
            args.append('--verbose')

        logger.info('exported functions:')
        for export in self.get_exported_functions():
            logger.info(f'\t{export}')
            args.append(f'-export={export}')

        myprocess.run_process(args)

    def _do_after_build(self):
        self._copy_to_output(self.file_output)
        self.file_output.unlink()
        self.file_ll.unlink()
        self.file_o.unlink()
        for ll_file in self.get_ll_files():
            try:
                ll_file.unlink()
            except FileNotFoundError:
                pass

    def _get_llvm_path(self):
        return dependencies.get_module_directory('llvm')

    def get_source_files(self):
        for filename in self.config['source_files']:
            yield (self.path / filename).expanduser().resolve()

    def get_ll_files(self):
        for source_file in self.get_source_files():
            yield source_file.with_suffix('.ll')

    def get_unit_file(self):
        first_file = next(self.get_source_files())
        return first_file

    def ensure_source_files(self):
        try:
            source_files = self.config['source_files']
            if len(source_files) == 0:
                source_files = self.get_source_files_from_folder()
        except KeyError:
            source_files = self.get_source_files_from_folder()

        self.config['source_files'] = source_files

    def get_exported_functions(self):
        file_export = self.find_file_globally('*.export')
        lines = utils.read_lines(file_export)
        return lines

    def default_config(self):
        config = super().default_config()
        config['language'] = 'clang'
        config['source_files'] = self.get_source_files_from_folder()
        return config

    def get_source_files_from_folder(self):
        return list(map(str, self.path.rglob('*.c')))

    def get_dependencies(self):
        return ['llvm']




from typing import Any, Dict, List, Tuple


class IAddress:
    def hex(self) -> str:
        return ""

    def bech32(self) -> str:
        return ""

    def pubkey(self) -> bytes:
        return bytes()


class IAccount:
    def get_seed(self) -> bytes:
        return bytes()


class ITransaction:
    def serialize(self) -> bytes:
        return bytes()

    def to_dictionary(self) -> Dict[str, Any]:
        return {}

    def to_dictionary_as_inner(self) -> Dict[str, Any]:
        return {}


class IElrondProxy:
    def get_account_nonce(self, address: IAddress) -> int:
        return 0

    def send_transaction(self, payload: Any) -> str:
        return ""

    def send_transactions(self, payload: List[Any]) -> Tuple[int, List[str]]:
        return 0, []

    def send_transaction_and_wait_for_result(self, payload: Any, num_seconds_timeout) -> str:
        return ""



import time

from erdpy import config
from erdpy.projects.templates_repository import TemplatesRepository


def get_templates_repositories():
    timestamp = int(time.time())
    examples_rs_tag = config.get_dependency_tag('elrond_wasm_rs')
    examples_rs_tag_no_v = examples_rs_tag[1:]

    return [
        TemplatesRepository(
            key="sc-examples",
            url=f"https://github.com/ElrondNetwork/sc-examples/archive/master.zip?t={timestamp}",
            github="ElrondNetwork/sc-examples",
            relative_path="sc-examples-master"
        ),

        TemplatesRepository(
            key="elrond-wasm-rs",
            url=f"https://github.com/ElrondNetwork/elrond-wasm-rs/archive/{examples_rs_tag}.zip?t={timestamp}",
            github="ElrondNetwork/elrond-wasm-rs",
            relative_path=f"elrond-wasm-rs-{examples_rs_tag_no_v}/contracts/examples"
        )
    ]



import setuptools

with open("README.md", "r") as fh:
    long_description = "https://github.com/ElrondNetwork/elrond-sdk-erdpy"

VERSION = "1.0.14"

try:
    with open('./erdpy/_version.py', 'wt') as versionfile:
        versionfile.write(f'__version__ = "{VERSION}"')
except FileNotFoundError:
    pass

# See https://packaging.python.org/tutorials/packaging-projects/
setuptools.setup(
    name="erdpy",
    version=VERSION,
    description="Elrond Smart Contracts Tools and Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ElrondNetwork/elrond-sdk-erdpy",
    author="Elrond Network",
    license="GPL",
    packages=setuptools.find_packages(
        include=["erdpy*"], exclude=["examples*"]),
    include_package_data=True,
    setup_requires=["wheel"],
    install_requires=[
        "toml>=0.10.2", "bottle", "requests", "pynacl", "pycryptodomex", "cryptography>=3.2", "prettytable"
    ],
    zip_safe=False,
    keywords=["Elrond"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: POSIX :: Linux",
        "Intended Audience :: Developers",
        "Development Status :: 3 - Alpha"
    ],
    entry_points={
        "console_scripts": [
            "erdpy=erdpy.cli:main",
        ],
    },
    python_requires=">=3.6"
)



import binascii


class Converters:
    @staticmethod
    def str_int_to_hex_str(number_str: str) -> str:
        num_of_bytes = 1
        if len(number_str) > 2:
            num_of_bytes = int(len(number_str) / 2)
        int_str = int(number_str)
        int_bytes = int_str.to_bytes(num_of_bytes, byteorder="big")
        bytes_str = binascii.hexlify(int_bytes).decode()
        return bytes_str

    @staticmethod
    def parse_keys(bls_public_keys):
        keys = bls_public_keys.split(',')
        parsed_keys = ''
        for key in keys:
            parsed_keys += '@' + key
        return parsed_keys, len(keys)



import itertools
import logging
import time
from pathlib import Path
from typing import Any, Dict, Iterator

import erdpy.config
import erdpy.utils as utils
from erdpy import workstation

logger = logging.getLogger("testnet")

ConfigurationType = Dict[str, Dict[str, Any]]


class Node:
    def __init__(self, index: int, folder: Path, shard: str, api_port: int) -> None:
        self.index = index
        self.folder = folder
        self.shard = shard
        self.api_port = api_port

    def config_folder(self):
        return self.folder / 'config'

    def key_file_path(self):
        return self.config_folder() / 'validatorKey.pem'

    def api_address(self):
        return f"http://localhost:{self.api_port}"

    def __repr__(self) -> str:
        return f"Node {self.index}, shard={self.shard}, port={self.api_port}, folder={self.folder}"


class TestnetConfiguration:

    def __init__(self, config):
        self.config = config

        sdk_folder = workstation.get_tools_folder()

        for key, path in self.config['folders'].items():
            if key == "testnet":
                continue

            path = path.replace('{ELRONDSDK}', str(sdk_folder))

            default_tag = erdpy.config.get_dependency_tag(key)
            path = path.replace('{TAG}', default_tag)

            # If the user has not specified a custom source repository, the
            # ones provided by the SDK will be used, which are downloaded as
            # tar.gz files from GitHub. Due to how the name of the archive is
            # built by GitHub, the path will contain the tag in two variants:
            # with the 'v' prefix (e.g. "v1.1.0"), but also without (e.g.
            # "1.1.0"), hence the need for {NOvTAG}.
            if default_tag.startswith("v"):
                default_tag = default_tag[1:]
            path = path.replace('{NOvTAG}', default_tag)

            self.config['folders'][key] = Path(path).expanduser()

        if 'testnet' not in self.config['folders']:
            self.config['folders']['testnet'] = Path().absolute() / 'testnet'

        self.__dict__.update(self.config)

    @classmethod
    def from_file(cls, filename):
        """
        If no filename is specified, try to load testnet.toml from the current
        directory, if there is any, and merge it with the SDK-level testnet
        config. Otherwise, load the SDK-level testnet config. If that one is
        missing as well, generate it from hard-coded defaults.
        """

        local_config = cls.get_local_config(filename)
        sdk_testnet_config = cls.get_sdk_testnet_config()
        final_config = merge_configs(sdk_testnet_config, local_config)
        return cls(final_config)

    @classmethod
    def from_sdk_testnet_config(cls):
        config = cls.get_sdk_testnet_config()
        return cls(config)

    @classmethod
    def from_default_config(cls):
        return cls(cls.default())

    @classmethod
    def get_local_config(cls, filename):
        if filename is None or filename == '':
            filename = Path() / "testnet.toml"
            if not filename.exists():
                return dict()

        return utils.read_toml_file(filename)

    @classmethod
    def get_sdk_testnet_config(cls):
        default = cls.default()
        filename = workstation.get_tools_folder() / "testnet.toml"

        if not filename.exists():
            logger.info('writing sdk_testnet_config from defaults')
            utils.write_toml_file(str(filename), default)
            return default
        sdk_testnet_config = utils.read_toml_file(filename)
        return merge_configs(default, sdk_testnet_config)

    def node_config_source(self):
        return self.node_source() / 'cmd' / 'node' / 'config'

    def node_source(self):
        return self.folders['elrond_go']

    def proxy_source(self):
        return self.folders['elrond_proxy_go']

    def proxy_config_source(self):
        return self.folders['elrond_proxy_go'] / 'cmd' / 'proxy' / 'config'

    def validator_key_files(self):
        for config_folder in self.validator_config_folders():
            yield config_folder / 'validatorKey.pem'

    def root(self):
        return self.folders['testnet']

    def genesis_time(self):
        return int(time.time()) + int(self.timing['genesis_delay'])

    def seednode_folder(self):
        testnet = self.root()
        return testnet / 'seednode'

    def seednode_config_folder(self):
        return self.seednode_folder() / 'config'

    def seednode_address(self):
        host = self.networking['host']
        port = self.networking['port_seednode']
        identifier = self.networking['p2p_id_seednode']
        return f"/ip4/{host}/tcp/{port}/p2p/{identifier}"

    def proxy_folder(self):
        testnet = self.root()
        return testnet / 'proxy'

    def proxy_config_folder(self):
        return self.proxy_folder() / 'config'

    def num_all_nodes(self):
        return self.num_all_validators() + self.num_all_observers()

    def num_all_validators(self):
        count = (
            self.shards['count'] * self.shards['validators_per_shard'] + self.metashard['validators']
        )
        return count

    def num_all_observers(self):
        count = (
            self.shards['count'] * self.shards['observers_per_shard'] + self.metashard['observers']
        )
        return count

    def num_shards(self):
        return self.shards['count']

    def num_observers_per_shard(self):
        return self.shards['observers_per_shard']

    def num_validators_per_shard(self):
        return self.shards['validators_per_shard']

    def num_validators_in_metashard(self):
        return self.metashard['validators']

    def all_nodes_folders(self):
        return itertools.chain(
            self.validator_folders(),
            self.observer_folders(),
        )

    def all_nodes_config_folders(self):
        return itertools.chain(
            self.validator_config_folders(),
            self.observer_config_folders(),
        )

    def validators(self) -> Iterator[Node]:
        first_port = self.networking['port_first_validator_rest_api']

        for i, folder in enumerate(self.validator_folders()):
            shard = self._get_shard_of_validator(i)
            port = first_port + i
            yield Node(index=i, folder=folder, shard=shard, api_port=port)

    def _get_shard_of_validator(self, observer_index: int):
        shard = int(observer_index // self.num_validators_per_shard())
        return shard if shard < self.num_shards() else self.metashard['metashardID']

    def validator_folders(self):
        testnet = self.root()
        for i in range(self.num_all_validators()):
            yield testnet / 'validator{:02}'.format(i)

    def validator_config_folders(self):
        for folder in self.validator_folders():
            yield folder / 'config'

    def observers(self) -> Iterator[Node]:
        first_port = self.networking['port_first_observer_rest_api']

        for i, folder in enumerate(self.observer_folders()):
            shard = self._get_shard_of_observer(i)
            api_port = first_port + i
            yield Node(index=i, folder=folder, shard=shard, api_port=api_port)

    def _get_shard_of_observer(self, observer_index: int):
        shard = int(observer_index // self.num_observers_per_shard())
        return shard if shard < self.num_shards() else self.metashard['metashardID']

    def observer_config_folders(self):
        for folder in self.observer_folders():
            yield folder / 'config'

    def observer_folders(self):
        testnet = self.root()
        for i in range(self.num_all_observers()):
            yield testnet / 'observer{:02}'.format(i)

    def observer_addresses(self):
        host = self.networking['host']
        first_port = self.networking['port_first_observer_rest_api']
        for port in range(self.num_all_observers()):
            port = first_port + port
            yield f"http://{host}:{port}"

    def validator_addresses(self):
        host = self.networking['host']
        first_port = self.networking['port_first_observer_rest_api']
        for port in range(self.num_all_observers()):
            port = first_port + port
            yield f"http://{host}:{port}"

    def api_addresses_sharded_for_proxy_config(self):
        nodes = list()
        for node in list(self.observers()):
            nodes.append({
                'ShardId': node.shard,
                'Address': node.api_address(),
                'Type': 'Observer'
            })

        for node in list(self.validators()):
            nodes.append({
                'ShardId': node.shard,
                'Address': node.api_address(),
                'Type': 'Validator'
            })

        return nodes

    def proxy_port(self):
        return self.networking["port_proxy"]

    def loglevel(self):
        return self.features.get("loglevel", "")

    def arwen_binary(self):
        return self.features.get("arwen_binary")


    @classmethod
    def default(cls):
        config = dict()
        config['features'] = {
            'loglevel': '*:DEBUG',
            'proxy': True,
            'arwen_binary': True
        }
        config['folders'] = {
            'elrond_go':
                '{ELRONDSDK}/elrond_go/{TAG}/elrond-go-{NOvTAG}',
            'elrond_proxy_go':
                '{ELRONDSDK}/elrond_proxy_go/{TAG}/elrond-proxy-go-{NOvTAG}',
        }
        config['metashard'] = {
            'consensus_size': 1,
            'metashardID': 4294967295,
            'observers': 0,
            'validators': 1,
        }
        config['networking'] = {
            'host': '127.0.0.1',
            'port_seednode': 9999,
            'p2p_id_seednode': '16Uiu2HAkw5SNNtSvH1zJiQ6Gc3WoGNSxiyNueRKe6fuAuh57G3Bk',
            'port_proxy': 7950,
            'port_first_observer': 21100,
            'port_first_observer_rest_api': 10000,
            'port_first_validator': 21500,
            'port_first_validator_rest_api': 10100,
        }
        config['shards'] = {
            'consensus_size': 1,
            'count': 1,
            'observers_per_shard': 0,
            'validators_per_shard': 1,
        }
        config['timing'] = {
            'genesis_delay': 10
        }

        return config


def merge_configs(leftcfg: ConfigurationType, rightcfg: ConfigurationType) -> ConfigurationType:
    result = dict()
    result.update(leftcfg)

    for key, subcfg in result.items():
        try:
            subcfg.update(rightcfg[key])
        except KeyError:
            pass

    for key in rightcfg:
        if key not in result:
            result[key] = rightcfg[key]

    return result



from erdpy.validators.core import (prepare_args_for_change_reward_address,
                                   prepare_args_for_claim,
                                   prepare_args_for_stake,
                                   prepare_args_for_unbond,
                                   prepare_args_for_unjail,
                                   prepare_args_for_unstake,
                                   prepare_args_for_unstake_nodes,
                                   prepare_args_for_unstake_tokens,
                                   prepare_args_for_unbond_tokens,
                                   prepare_args_for_unbond_nodes,
                                   prepare_args_for_clean_registered_data,
                                   prepare_args_for_restake_unstaked_nodes)

from erdpy.validators.validators_file import ValidatorsFile

__all__ = ["prepare_args_for_stake",
           "prepare_args_for_unstake",
           "prepare_args_for_unbond",
           "prepare_args_for_unjail",
           "prepare_args_for_change_reward_address",
           "prepare_args_for_claim",
           "prepare_args_for_unstake_nodes",
           "prepare_args_for_unstake_tokens",
           "prepare_args_for_unbond_nodes",
           "prepare_args_for_unbond_tokens",
           "prepare_args_for_clean_registered_data",
           "prepare_args_for_restake_unstaked_nodes",
           "ValidatorsFile"]



VM_TYPE_SYSTEM = "0001"
VM_TYPE_ARWEN = "0500"
SC_HEX_PUBKEY_PREFIX = "0" * 16
SC_HEX_PUBKEY_PREFIX_SYSTEM = SC_HEX_PUBKEY_PREFIX + VM_TYPE_SYSTEM + "0" * 30
SC_HEX_PUBKEY_PREFIX_ARWEN = SC_HEX_PUBKEY_PREFIX + VM_TYPE_ARWEN



from typing import Any

from erdpy.testnet import genesis
from erdpy.testnet.config import TestnetConfiguration


def patch(data: Any, testnet_config: TestnetConfiguration):
    owner = genesis.get_owner_of_genesis_contracts()

    delegation_config = data[0]
    dns_config = data[1]

    delegation_config["owner"] = owner.address.bech32()
    dns_config["owner"] = owner.address.bech32()
    # registration price = 100 atoms of eGLD
    dns_config["init-parameters"] = "0064"



import logging

from erdpy.blockatlas.http_facade import do_get

logger = logging.getLogger("block-atlas")


class BlockAtlas:
    def __init__(self, url, coin):
        self.url = url
        self.coin = coin

    def get_current_block_number(self):
        pass

    def get_block_by_number(self, num: int):
        pass

    def get_txs_by_address(self, address: str):
        url = f"{self.url}/v2/{self.coin}/transactions/{address}"
        response = do_get(url)
        return response




import logging
import sys

import requests

from erdpy import errors

logger = logging.getLogger("downloader")

CHUNK_SIZE = 1024 * 16
LINECLEAR = '\r' + ' ' * 20 + '\r'

PROGRESS_RULER = 'Downloading...\n|_,_,_,_,_,,_,_,_,_,_|'


def download(url: str, filename: str) -> None:
    if url is None:
        raise errors.BadUrlError()

    logger.info(f"download_url.url: {url}")
    logger.info(f"download_url.filename: {filename}")

    try:
        response = requests.get(url, stream=True)
        response.raise_for_status()
        print(PROGRESS_RULER, file=sys.stderr)
        print(' ', end='', file=sys.stderr)
        sys.stderr.flush()
        total_size = int(response.headers.get("content-length", 0))
        chunk_number = 0
        progress = 0
        with open(filename, 'wb') as file:
            for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                file.write(chunk)
                progress = _report_download_progress(progress, chunk_number, total_size)
                chunk_number += 1
        print('', file=sys.stderr)
        sys.stderr.flush()
    except requests.HTTPError as err:
        raise errors.DownloadError(
            f"Could not download [{url}] to [{filename}]") from err

    logger.info("Download done.")


def _report_download_progress(progress, chunk_number, total_size):
    try:
        num_chunks = int(total_size / CHUNK_SIZE)
        new_progress = int((chunk_number / num_chunks) * 20)
        if new_progress > progress:
            progress_markers = '·' * (new_progress - progress)
            print(progress_markers, end='', file=sys.stderr)
        sys.stderr.flush()
        return new_progress
    except ZeroDivisionError:
        return 0
